# Responses
RESPONSE_OK = "1"
RESPONSE_ERROR = "0"

# Statuses
STATUS_DISCONNECTED = "Disconnected"
STATUS_CONNECTING = "Connecting"
STATUS_CONNECTED = "Connected"
STATUS_CONFIGURING = "Configuring"
STATUS_CONFIGURED = "Configured"
STATUS_IDLE = "Idle"
STATUS_BUSY = "Busy"
STATUS_ERROR = "Error"
STATUS_DISCONNECTING = "Disconnecting"

READY_DEVICE_STATUSES = (STATUS_CONFIGURED, STATUS_IDLE, STATUS_BUSY)
IDLE_DEVICE_STATUSES = (STATUS_CONFIGURED, STATUS_IDLE)
STATUSES_PRIORITY = (
    STATUS_ERROR, STATUS_DISCONNECTING, STATUS_DISCONNECTED, STATUS_CONNECTING, STATUS_CONNECTED, STATUS_CONFIGURING,
    STATUS_BUSY, STATUS_IDLE, STATUS_CONFIGURED)

# Macro states
IDLE = 0
BUSY = 1
PAUSED = 2
DONE = 3
STOPPED = 4
ERROR = 5
WARNING = 6

ABORT = "Abort macro"
SKIP = "Skip"

MODE_ALL = 2
MODE_EXTENDED = 1
MODE_SIMPLE = 0

# Attribute values
VALUE = "Value"
TYPE = "Type"
MIN = "Min"
MAX = "Max"
UNIT = "Unit"
DECIMALS = "Decimals"
DESCRIPTION = "Description"
READONLY = "Readonly"
TYPE_LIST = "<type> List"
TYPE_LIST_VALUES = "<type> List values"
TYPE_ARRAY = "<type> Array"
FACTOR = "Factor"
OFFSET = "Offset"
TOLERANCE = "Tolerance"
SETPOINT = "Setpoint"
PREVIOUS = "Previous"
SAVE_CHANGE = "Save"
ALL = "*"
DISPLAY = "display"
IMAGE = "Image"
LOCAL_FILESYSTEM = "Local filesystem"
ALLOW_BROADCAST = "Allow broadcast"

# Basic attributes
ATTR_ID = "Id"
ATTR_DESCRIPTION = "Description"
ATTR_STATUS = "Status"
ATTR_ENABLED = "Enabled"

INSTRUCTION_GET = b"0"
INSTRUCTION_SET = b"1"
INSTRUCTION_COMMAND = b"2"
INSTRUCTION_INIT = b"3"
INSTRUCTION_METHOD = b"4"
INSTRUCTION_POLL = b"5"
INSTRUCTION_PULL = b"6"
INSTRUCTION_PUSH = b"7"
INSTRUCTION_FORWARD = b"8"
INSTRUCTION_SYNC = b"9"

# Tokens
TOKEN_ATTRIBUTE = "A!"
TOKEN_ATTRIBUTE_REPLACE = "R!"
TOKEN_ATTRIBUTE_DELETE = "D!"
TOKEN_ATTRIBUTE_GROUP_DELETE = "GD!"
TOKEN_ALL = "*"

# Words
RISING = u"Rising"
FALLING = u"Falling"
POSITIVE = u"Positive"
NEGATIVE = u"Negative"
ROWS = u"Rows"
COLUMNS = u"Columns"

# Macro steps
MACRO_SET_ATTRIBUTE_STEP = u"Set attribute"
MACRO_EXECUTE_METHOD_STEP = u"Execute method"

# Config enums
CFG_TEMPLATE = 0

# Zmq message tokens
WORKER_READY = b"1"
MSG_TERMINATE = b"0"
WORKER_DONE = b"2"
MSG_EMPTY = b""
MSG_JSON = b"J"
MSG_ARRAY = b"A"
MSG_BYTES = b"B"
MSG_PICKLE = b"P"
MSG_FILE = b"F"
MSG_ACK = b"ACK"
MSG_FORWARD = b"FRWD"
MSG_ABORT = b"ABT"
MSG_PROGRESS = b"PROG"

OPEN_LOOP = "Open loop"
CLOSE_LOOP = "Close loop"

PROXY = "PROXY"

# Session groups
GROUP_RC = u"RocketChat"

# List of general Attributes
ATTR_SERIAL_NUMBER = u"Serial number"
ATTR_PRODUCT_TYPE = u"Product type"
ATTR_MODEL = u"Model name"
ATTR_ANALOG_INPUT_READOUT_TASK = u"Analog readout task"
ATTR_TERMINAL_CONFIGURATION = u"Terminal configuration"
ATTR_SAMPLES_NUMBER = u"Number of samples"
ATTR_READOUT_AT_SAMPLES = u"Readout at samples"
ATTR_TEMPERATURE = u"Temperature"
ATTR_CHANNELS = u"Channels"
ATTR_POSITION = u"Position"
ATTR_TARGET_POSITION = u"Target position"
ATTR_STALL_FLAG = u"Stall flag"
ATTR_MOVING = u"Moving"
ATTR_VELOCITY = u"Velocity"
ATTR_HOLDING_CURRENT = u"Holding current"
ATTR_MAXIMUM_VELOCITY = u"Maximum velocity"
ATTR_RUN_CURRENT = u"Run current"
ATTR_ROTATION_DIRECTION = u"Rotation direction"
ATTR_TORQUE_SPEED = u"Torque speed"
ATTR_ENCODER_ENABLED = u"Encoder enabled"
ATTR_MICROSTEP_RESOLUTION = u"Microstep resolution"
ATTR_STEPS_PER_REVOLUTION = u"Steps / revolution"
ATTR_STALL_FACTOR = u"Stall factor"
ATTR_POSITIVE_LIMIT_SWITCH = u"+limit switch"
ATTR_NEGATIVE_LIMIT_SWITCH = u"-limit switch"
ATTR_INDEX_MARK = u"Index mark"
ATTR_USE_POSITIVE_LIMIT_SWITCH = u"Use +limit switch"
ATTR_USE_NEGATIVE_LIMIT_SWITCH = u"Use -limit switch"
ATTR_INITIAL_VELOCITY = u"Initial velocity"
ATTR_REFERENCING = u"Referencing"
ATTR_ERROR_CODE = u"Error code"
ATTR_LATENCY = u"Latency"
ATTR_MODULES = u"Modules count"
ATTR_STATE = u"State"
ATTR_ENCODER_BASE_UNIT = u"Enc base unit"
ATTR_HAND_TOOL = u"Hand tool"
ATTR_MOVEMENT_LOCKED = u"Movement locked"
ATTR_INTERNAL_COMM_FAILURE = u"Comm failure"
ATTR_STREAMING = u"Streaming"
ATTR_HOLDING = u"Holding"
ATTR_SENSOR_PRESENT = u"Sensor present"
ATTR_BOOSTER_PRESENT = u"Booster present"
ATTR_ADJUSTMENT_ACTIVE = u"Adjustment active"
ATTR_IOM_PRESENT = u"I/O present"
ATTR_HIGH_VOLTAGE_FAILURE = u"High voltage failure"
ATTR_HIGH_VOLTAGE_OVERLOAD = u"High voltage overload"
ATTR_OVER_TEMPERATURE = u"Over temperature"
ATTR_CLOSED_LOOP_ACTIVE = u"Closed loop active"
ATTR_CALIBRATING = u"Calibrating"
ATTR_MOVE_DELAYED = u"Move delayed"
ATTR_CALIBRATED = u"Calibrated"
ATTR_REFERENCED = u"Referenced"
ATTR_END_STOP_REACHED = u"On end stop"
ATTR_RANGE_LIMIT_REACHED = u"On range limit"
ATTR_FOLLOWING_LIMIT_REACHED = u"On Following limit"
ATTR_MOVEMENT_FAILED = u"Movement failed"
ATTR_POSITIONER_TYPE = u"Positioner type"
ATTR_POSITIONER_CODE = u"Positioner code"
ATTR_POSITION_UPPER_LIMIT = u"Position +limit"
ATTR_POSITION_LOWER_LIMIT = u"Position -limit"
ATTR_STEP_FREQUENCY = u"Step frequency"
ATTR_STEP_AMPLITUDE = u"Step amplitude"
ATTR_ACCELERATION = u"Acceleration"
ATTR_DECELERATION = u"Deceleration"
ATTR_SENSOR_MODE = u"Sensor mode"
ATTR_MOVE_MODE = u"Move mode"
ATTR_SAFE_DIRECTION = u"Safe direction"
ATTR_HAND_TOOL_MODE = u"Hand tool mode"
ATTR_SENSOR_TYPE = u"Sensor type"
ATTR_REVOLUTIONS = u"Revolutions"
ATTR_BAUDRATE = u"Baudrate"
ATTR_REFERENCING_METHOD = u"Referencing method"
ATTR_ACCELERATION_PATTERN = u"Acceleration pattern"
ATTR_ACCELERATION_TIME = u"Acceleration time"
ATTR_CORRECTION = u"Position correction"
ATTR_CORRECTION_RETRY = u"Correction retry_count"
ATTR_DECELERATION_TIME = u"Deceleration time"
ATTR_ENCODER_ADDING = u"Encoder adding"
ATTR_ENCODER_READ = u"Encoder read"
ATTR_ENCODER_RESOLUTION = u"Encoder resolution"
ATTR_PULSE_POSITION = u"Pulse position"
ATTR_PULSE_READ = u"Pulse read"
ATTR_START_SPEED = u"Start speed"
ATTR_STEP_FACTOR = u"Step factor"
ATTR_STEP_RESOLUTION = u"Step resolution"
ATTR_TOP_SPEED = u"Top speed"
ATTR_TRIGGER_EDGE = u"Trigger edge"
ATTR_TRIGGER_ENC_PITCH = u"Trigger enc pitch"
ATTR_TRIGGER_LOGIC = u"Trigger logic"
ATTR_TRIGGER_PM_PITCH = u"Trigger pm pitch"
ATTR_TRIGGER_PULSE_WIDTH = u"Trigger pulse width"
ATTR_TRIGGER_SELECT = u"Trigger select"
ATTR_TRIGGER_SOURCE = u"Trigger source"
ATTR_ENCODER_CALC_NUM = u"Encoder calc num"
ATTR_ENCODER_CALC_DEN = u"Encoder calc den"
ATTR_ROTATION_CHANGE = u"Rotation change"
ATTR_SWAP_LIMITS = u"Swap limits"
ATTR_RETRY_COUNT = u"Retry count"
ATTR_FEEDBACK_TIMEOUT = u"Feedback timeout"
ATTR_FEEDBACK_CONTROL_TYPE = u"Feedback type"
ATTR_MOTOR_EXCITATION = u"Motor excitation"
ATTR_MOTOR_SELECTION = u"Motor selection"
ATTR_MICROSTEP_SELECT = u"Microstep select"
ATTR_STOP_TYPE = u"Stop type"
ATTR_EMERGENCY_STOP = u"Emergency stop"
ATTR_CWL = u"CW limit"
ATTR_CCWL = u"CCW limit"
ATTR_ORG = u"ORG signal"
ATTR_NORG = u"NORG signal"
ATTR_LAST_ERROR = u"Last error"
ATTR_LOG_DIRECTORY = u"Log directory"
ATTR_SCAN_DIRECTORY = u"Scan directory"
ATTR_ALLOW_CONSOLE_LOG = u"Allow console log"
ATTR_ALLOW_DEVICE_LOG = u"Allow device log"
ATTR_ALLOW_GUI_LOG = u"Allow gui log"
ATTR_ALLOW_ATTRIBUTE_LOG = u"Allow attribute log"
ATTR_ALLOW_MACRO_LOG = u"Allow macro log"
ATTR_DEVICES_COUNT = u"Devices count"
ATTR_LOOPS_RUNNING = u"Control loops running"
ATTR_FILEPATH = u"Filepath"
ATTR_CONTENT = u"Content"
ATTR_STEP_INDEX = u"Step index"
ATTR_POINT_INDEX = u"Point index"
ATTR_STEPS_COUNT = u"Steps count"
ATTR_POINTS_COUNT = u"Points count"
ATTR_PROGRESS = u"Progress"
ATTR_STARTED_AT = u"Started at"
ATTR_RUNNING_TIME = u"Running time"
ATTR_ESTIMATED_TIME = u"Estimated time"
ATTR_ESTIMATED_END = u"Estimated end"
ATTR_REMAINING_TIME = u"Remaining time"
ATTR_CHAIN_LINK = u"Chain link"
ATTR_OPENED = u"Opened"
ATTR_OPENED_POSITION = u"Opened position"
ATTR_CLOSED_POSITION = u"Closed position"
ATTR_OPENED_VALUE = u"Opened value"
ATTR_CLOSED_VALUE = u"Closed value"
ATTR_OPENED_AT = u"Opened at"
ATTR_AUTO_MODE = u"Auto mode"
ATTR_SYNTAX_VERSION = u"Syntax vesrion"
ATTR_SERVO_STATE = u"Servo state"
ATTR_ON_TARGET = u"On target"
ATTR_ACTIVE_UNIT = u"Active unit"
ATTR_REFERENCING_DIRECTION_X = u"Ref direction X"
ATTR_REFERENCING_DIRECTION_Y = u"Ref direction Y"
ATTR_REFERENCING_DIRECTION_Z = u"Ref direction Z"
ATTR_PIVOT_MODE = u"Pivot mode"
ATTR_PIVOT_X = u"Pivot X"
ATTR_PIVOT_Y = u"Pivot Y"
ATTR_PIVOT_Z = u"Pivot Z"
ATTR_POSITIONS = u"Positions"
ATTR_USE_MARK_INDEX = u"Use mark index"
ATTR_BACKLASH_DISTANCE = u"Backlash distance"
ATTR_BUFFERED_COMMANDS = u"Buffered commands"
ATTR_HANGING_COMMANDS = u"Hanging commands"
ATTR_CONSTANT_CURRENT = u"Constant current"
ATTR_CONSTANT_VOLTAGE = u"Constant voltage"
ATTR_CURRENT = u"Current"
ATTR_VOLTAGE = u"Voltage"
ATTR_MAX_CURRENT = u"Max current"
ATTR_MAX_VOLTAGE = u"Max voltage"
ATTR_OUTPUT_SWITCH = u"Output switch"
ATTR_INTENSITY = u"Intensity"
ATTR_RESOURCE_DIRECTORY = u"Resource directory"
ATTR_CACHE_DIRECTORY = u"Cache directory"
ATTR_SCAN_PREFIX = u"Scan prefix"
ATTR_SCAN_COUNT = u"Scan count"
ATTR_OUTPUT_FILENAME = u"Output filename"
ATTR_TIMEOUT = u"Timeout"
ATTR_ACRONYM = u"Acronym"
ATTR_MOTOR_TYPE = u"Motor type"
ATTR_AMPLIFIER_GAIN = u"Amplifier gain"
ATTR_FAST_REFERENCE_SPEED = u"Fast reference speed"
ATTR_FULL_STEP_RESOLUTION = u"Full step resolution"
ATTR_HOMING_DIRECTION = u"Homing direction"
ATTR_JOG_BACKWARD = u"Jog backward"
ATTR_JOG_FORWARD = u"Jog forward"
ATTR_LATCH_INPUT_LEVEL = u"Latch input level"
ATTR_LIMIT_SWITCH_ACTIVE_LEVEL = u"Limit switch active level"
ATTR_LOW_CURRENT_STEPPER_MODE = u"Low current stepper mode"
ATTR_SLOW_REFERENCE_SPEED = u"Slow reference speed"
ATTR_STEP_DRIVE_RESOLUTION = u"Step drive resolution"
ATTR_JOG_MOTOR = u"Jog motor"
ATTR_HOMING_METHOD = u"Homing method"
ATTR_LIMIT_SWITCHES_MODE = u"Limit switches mode"
ATTR_REMOTE_TIMESTAMP = u"Remote timestamp"
ATTR_LOCAL_TIMESTAMP = u"Local timestamp"
ATTR_DELTA_TIMESTAMP = u"Delta timestamp"
ATTR_REMOTE_TRAIN_ID = u"Remote train id"
ATTR_LOCAL_TRAIN_ID = u"Local train id"
ATTR_MIN_UPDATE_TIMEOUT = u"Min update timeout"
ATTR_TRAIN_TIMEOUT = u"Train timeout"
ATTR_TRAIN_OFFSET = u"Train id offset"
ATTR_COUNTER = u"Counter"
ATTR_REFRESHING = u"Refreshing"
ATTR_WAVETABLE_LENGTH = u"Wavetable length"
ATTR_WAVETABLE_DATA = u"Wave table data"
ATTR_WAVETABLE_RATE = u"Wave table rate"
ATTR_WAVETABLE_INTERPOLATION = u"Wave table interpolation"
ATTR_WAVE_GEN_OFFSET = u"Wave gen offset"
ATTR_TRIGGER_OUTPUTS_COUNT = u"Trigger outputs count"
ATTR_WAVE_GEN_COUNT = u"Wave generators count"
ATTR_WAVE_TABLE = u"Wave table"
ATTR_SERVO_UPDATE_TIME = u"Servo update time"
ATTR_RECORD_TABLE_RATE = u"Record table rate"
ATTR_RECORD_TABLE_DATA = u"Record table data"
ATTR_RECORD_TABLE_DATA_SETPOINTS = u"Record table data setpoints"
ATTR_WAVETABLE_OFFSET = u"Wave table offset"
ATTR_TRIGGER_LINE_DATA = u"Trigger line data"
ATTR_SCANNING = u"Scanning"
ATTR_READOUT = u"Readout"
ATTR_CAMERA_CALIBRATION = u"Pixel calibration"
ATTR_COLOR_OUTPUT = u"Color output"
ATTR_COMMENT = u"Comment"
ATTR_DUMMY_IMAGE = u"Dummy image"
ATTR_EXPOSURE_TIME = u"Exposure time"
ATTR_FRAMERATE = u"Framerate"
ATTR_FRAME_IMAGE = u"Frame image"
ATTR_GAIN = u"Gain"
ATTR_HEIGHT_DIVIDER = u"Height divider"
ATTR_HEIGHT_MAX = u"Height max"
ATTR_LIVE_VIEW = u"Live view"
ATTR_LIVE_VIEW_FPS_LIMIT = u"Live view limit"
ATTR_OUTPUT_NAME = u"Output name"
ATTR_PIXEL_BITSIZE = u"PixelSize"
ATTR_REAL_FRAMERATE = u"Real framerate"
ATTR_ROI_OFFSET_X = u"ROI offset X"
ATTR_ROI_OFFSET_Y = u"ROI offset Y"
ATTR_SAVE_FRAMES = u"Save frames"
ATTR_START_ACQUISITION = u"Start acquisition"
ATTR_STOP_ACQUISITION = u"Stop acquisition"
ATTR_SUMMED_FRAMES = u"Summed frames"
ATTR_TRIGGER_MODE = u"Trigger mode"
ATTR_WIDTH_DIVIDER = u"Width divider"
ATTR_WIDTH_MAX = u"Width max"
GROUP_ROI = u"Roi"
ATTR_WAVE_GENERATOR = u"Wave generator"
ATTR_OUTPUT_DIRECTORY = u"Output directory"
ATTR_TRIGGER_DURATION = u"Trigger duration"
ATTR_SETPOINT = u"Setpoint"
ATTR_FAULT_BYTE = u"Fault byte"
ATTR_FAST_REFERENCE_VELOCITY = u"Fast reference velocity"
ATTR_SLOW_REFERENCE_VELOCITY = u"Slow reference velocity"
ATTR_REFERENCE_STEPS_BACK = u"Reference steps back"
ATTR_HOLDING_TIME = u"Holding time"
ATTR_STEPS = u"Steps"
ATTR_PRESSURE = u"Pressure"
ATTR_POWER = u"Power"
ATTR_CURRENT_SAMPLE = u"Current sample"
ATTR_CURRENT_CHANNEL = u"Current channel"
ATTR_ACTIVE_SWITCHES = u"Active switches"
ATTR_LIBRARY = u"Library"
ATTR_ENCODER_READOUT = u"Encoder readout"
ATTR_COUNTS_PER_REVOLUTION = u"Counts pre revolution"
ATTR_REFERENCE_VELOCITY = u"Reference velocity"
ATTR_DIRECTION = u"Direction"
ATTR_DWELL_TIME = u"Dwell time"
ATTR_TEMPLATES_COUNT = u"Templates count"
ATTR_MACRO_COUNT = u"Macro count"
ATTR_MACRO_PREFIX = u"Macro prefix"
ATTR_TEMPLATE_DIRECTORY = u"Template directory"
ATTR_UPLOADED_COUNT = u"Uploaded count"
ATTR_FINISHED_COUNT = u"Finished count"
ATTR_RUNNING_COUNT = u"Running count"
ATTR_MAX_RUNNING_MACROS = u"Max running macros"
ATTR_MAX_MACRO_HISTORY = u"Max macro history"
ATTR_TEMPLATE = u"Template"
ATTR_CHAIN_LINK_OUTPUT = u"Chain link output"
ATTR_OFF = u"Off"
ATTR_PS1_VOLTAGE = u"PS 1 voltage"
ATTR_PS1_POWER = u"PS 1 power"
ATTR_PS1_CURRENT = u"PS 1 current"
ATTR_PS2_VOLTAGE = u"PS 2 voltage"
ATTR_PS2_POWER = u"PS 2 power"
ATTR_PS2_CURRENT = u"PS 2 current"
ATTR_PS3_VOLTAGE = u"PS 3 voltage"
ATTR_PS3_POWER = u"PS 3 power"
ATTR_PS3_CURRENT = u"PS 3 current"
ATTR_PS4_VOLTAGE = u"PS 4 voltage"
ATTR_PS4_POWER = u"PS 4 power"
ATTR_PS4_CURRENT = u"PS 4 current"
ATTR_POWER_SETPOINT = u"Power setpoint"
ATTR_VOLTAGE_SETPOINT = u"Voltage setpoint"
ATTR_CURRENT_SETPOINT = u"Current setpoint"
ATTR_OUTPUT_RELEASED = u"Output released"
ATTR_PROCESS_ID = u"Process ID"
ATTR_MEMORY_USAGE = u"Memory usage"
ATTR_CPU_USAGE = u"Cpu usage"
ATTR_AVAILABLE_MEMORY = u"Available memory"
ATTR_HOME_SWITCH_FLAG = u"Home switch flag"
ATTR_HARDWARE_VERSION = u"Hardware version"
ATTR_SOFTWARE_VERSION = u"Software version"
ATTR_MANUFACTURE_DATE = u"Manufacture date"
ATTR_APPLICATION_NUMBER = u"Application number"
ATTR_APPLICATION_VERSION = u"Application version"
ATTR_OPERATION_MODE = u"Operation mode"
ATTR_VELOCITY_UNITS = u"Velocity units"
ATTR_WIDTH = u"Width"
ATTR_HEIGHT = u"Height"
ATTR_INPUT_DIR_LIST = u"Input dir list"
ATTR_OUTPUT_DIR_LIST = u"Output dir list"
ATTR_FILE_FILTER_PATTERN = u"File filter pattern"
ATTR_ADDRESS = u"Address"
ATTR_HOST = u"Host"
ATTR_PORT = u"Port"
ATTR_TASKS = u"Tasks"
ATTR_WORKER = u"Worker"
ATTR_CLIENT = u"Client"
ATTR_TASK_INSTRUCTION = u"Task instruction"
ATTR_SOURCE_FILE = u"Source file"
ATTR_TARGET_FILE = u"Target file"
ATTR_ABORTED = u"Aborted"
ATTR_DRIVES = u"Drives"
ATTR_FINISHED_AT = u"Finished at"
ATTR_REQUESTED_FILES = u"Requested files"
ATTR_SYNCED_FILES = u"Synced files"
ATTR_METADATA = u"Metadata"
ATTR_LOGFILE = u"Logfile"
ATTR_HEADER = u"Header"
ATTR_LAST_LOG_LINE = u"Last log line"
ATTR_INTERVAL = u"Interval"
ATTR_LOGGING = u"Logging"
ATTR_MEASURING_SCALE = u"Measuring scale"
ATTR_DISTANCE = u"Distance"
ATTR_MODES = u"Modes"
ATTR_AUTO_STOP = u"Auto stop"
ATTR_ABSOLUTE_POSITION = u"Absolute position"
ATTR_WARNING = u"Warning"
ATTR_VELOCITY_TIMEOUT = u"Velocity timeout"
ATTR_SPEED_TABLE_ROW = u"Speed table row"
ATTR_RESISTANCE = u"Resistance"
ATTR_ALLOW_RC = u"Allow RocketChat"
ATTR_MAIN_ROOM = u"Main room name"
ATTR_RC_MACRO_DISCUSSION = u"Macro ds name"
ATTR_AUTH_TOKEN = u"Auth token"
ATTR_USER_NAME = u"username"
ATTR_PASSWORD = u"Password"

ATTR_ANALOGUE_PROGRAMMING_ACTIVE = u"Analogue prog active"
ATTR_CALIBRATION_MODE = u"Calibration mode"
ATTR_CONTROL_CURRENT = u"Control current"
ATTR_CONTROL_VOLTAGE = u"Control voltage"
ATTR_CURRENT_RAMPING_RATE = u"Current ramping rate"
ATTR_DIGITAL_PROGRAMMING_ACTIVE = u"Digital prog active"
ATTR_MONITORED_CURRENT = u"I-Mon"
ATTR_MONITORED_VOLTAGE = u"V-Mon"
ATTR_OUTPUT_POLARITY = u"Output polarity"
ATTR_OUTPUT_RELEASE = u"Output release"
ATTR_POLARITY_REVERSAL = u"Polarity"
ATTR_VOLTAGE_RAMPING_RATE = u"Voltage ramping rate"
ATTR_VOLTAGE_READOUT = u"Voltage readout"
ATTR_CURRENT_READOUT = u"Current readout"
ATTR_CURRENT_ON_TARGET = u"Current on target"
ATTR_VOLTAGE_ON_TARGET = u"Voltage on target"
ATTR_RAMP_VOLTAGE_MODE = u"Ramp voltage mode"
ATTR_RAMP_CURRENT_MODE = u"Ramp current mode"
ATTR_OUTPUT_POWER = u"Output power"
ATTR_DIGITAL_OUTPUT = "Digital output"
ATTR_DIGITAL_INPUT = "Digital input"
ATTR_ANALOG_OUTPUT = "Analog output"
ATTR_ANALOG_INPUT = "Analog input"
ATTR_CHANNELS_CONFIGURATION = "Channels configuration"
ATTR_NAME = "Name"
ATTR_MEASUREMENT_STATUS = "Measurement status"
ATTR_FIRMWARE = "Firmware version"
ATTR_SENSORS_COUNT = "Sensors count"
GROUP_MATERIALS = "Materials"
ATTR_MATERIAL = "Material"
ATTR_AUTO_Z = "Auto Z"
ATTR_OUTPUT = "Output"
GROUP_OUTPUTS = "Outputs"
ATTR_PROCESS = "Process"
GROUP_PROCESSES = "Processes"
GROUP_USER_MESSAGES = "User messages"
ATTR_USER_MESSAGE = "User message"
ATTR_Z_RATIO = "Z ratio"
ATTR_FUNDAMENTAL_FREQUENCY = "Fundamental freq"
ATTR_AHARMONIC_FREQUENCY = "Aharmonic freq"
ATTR_SENSOR_THICKNESS = "Sensor thickness"
ATTR_CRYSTAL_LIFE = "Crystal life"
ATTR_RAW_RATE = "Raw rate"
ATTR_SHUTTER_STROKES = "Shutter strokes"
ATTR_SHUTTER_OPEN = "Shutter open"
ATTR_LAYER_THICKNESS = "Layer thickness"
ATTR_ACTIVE_PROCESS = "Active process"
ATTR_GROUPS = "Groups"
ATTR_LAST_UPDATE = "Last update"
ATTR_SOURCE_ID = "Source id"
GROUP_FRAME = "Frame"
ATTR_FRAME = "Frame"
ATTR_FRAME_COUNT = "Frame count"
ATTR_ALLOW_STREAM = "Allow stream"
ATTR_MAX_STREAM_RATE = "Max stream rate"
ATTR_ALLOW_BROADCAST = "Allow broadcast"
ATTR_MAX_BROADCAST_RATE = "Max broadcast rate"
ATTR_SOURCE = "Source"
ATTR_SOURCE_FRAMERATE = "Source framerate"
ATTR_SOURCE_WIDTH = "Source width"
ATTR_SOURCE_HEIGHT = "Source height"
ATTR_AUTO_SET_FRAMERATE = "Autoset framerate"
ATTR_CAPTURE_API = "Capture API"
ATTR_SOURCES_COUNT = "Sources count"
ATTR_STREAM_READOUT = "Stream readout"
ATTR_VALUE_READOUT = "Value readout"
ATTR_STREAMABLE = "Streamable"
ATTR_FULL_ACCESS = "Full access"
ATTR_PILOT_LASER = "Pilot laser"
ATTR_IS_STABLE = "Is stable"
ATTR_CAPTURE_REGION = "Capture region"
ATTR_FourCC_CODEC = "FourCC Codec"
ATTR_FILENAME_PREFIX = "Filename prefix"
ATTR_RECORDING = "Recording"
ATTR_FRAME_AGGREGATION = "Frame aggregation"
ATTR_BUFFER_AGGREGATION = "Buffer aggregation"
ATTR_STREAMING_MODE = "Streaming mode"
ATTR_BUFFERS_INTERLEAVED = "Buffers interleaved"
ATTR_FILTER_RATE = "Filter rate"
ATTR_BUFFERS_COUNT = "Buffers count"
ATTR_FRAMES_GRABBED = "Frames grabbed"
ATTR_HEAD_FIBER_LENGTH = "Head fiber length"
ATTR_EXTENSION_FIBER_LENGTH = "Extension fiber length"
ATTR_HEAD_TYPE = "Head type"
ATTR_MIN_WORKING_DISTANCE = "Min working distance"
ATTR_MAX_WORKING_DISTANCE = "Max working distance"
ATTR_BEAM_INTERRUPT_TOLERANCE = "Beam interrupt tolerance"
ATTR_CLOSE_LOOP_UNIT = "Close loop unit"
ATTR_CLOSE_LOOP_FACTOR = "Close loop factor"
ATTR_MOVE_METHOD = "Move method"
ATTR_CHANNEL_ENABLED = "Channel enabled"


import os
import re
import time
from collections import OrderedDict
from datetime import datetime
from threading import RLock

import numpy as np
import oyaml
import yaml

import kamzik3
from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.macro.common import StepGenerator
from kamzik3.macro.scan import Scan
from kamzik3.macro.step import StepDeviceAttributeNumerical
from kamzik3.snippets.snippetDataAccess import get_next_file_index
from kamzik3.snippets.snippetsDecorators import expose_method


class MacroServer(Device):
    factory_class_names = {
        "StepDeviceAttributeNumerical": StepDeviceAttributeNumerical
    }

    def __init__(self, device_id=None, config=None):
        self.loaded_templates = {}
        self.uploaded_macros = OrderedDict()
        self.running_macros = OrderedDict()
        self.finished_macros = OrderedDict()
        self.update_lock = RLock()
        super().__init__(device_id, config)
        self.connect()

    def handle_configuration(self):
        start_at = time.time()
        self._config_commands()
        self._config_attributes()
        self.set_macro_log_directory()
        self.set_macro_templates_directory()
        self.get_macro_count()
        self.read_scan_templates_directory()
        self.set_status(STATUS_CONFIGURED)
        self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

    def _init_attributes(self):
        Device._init_attributes(self)
        self.create_attribute(ATTR_TEMPLATES_COUNT, default_value=0, readonly=True)
        self.create_attribute(ATTR_MACRO_COUNT, default_value=0, default_type=np.uint64, readonly=True)
        self.create_attribute(ATTR_MACRO_PREFIX, default_value=u"Scan", set_function=self.set_macro_prefix,
                              set_value_when_set_function=False)
        self.create_attribute(ATTR_UPLOADED_COUNT, default_value=0, default_type=np.uint64, readonly=True)
        self.create_attribute(ATTR_RUNNING_COUNT, default_value=0, default_type=np.uint64, readonly=True)
        self.create_attribute(ATTR_FINISHED_COUNT, default_value=0, default_type=np.uint64, readonly=True)
        self.create_attribute(ATTR_OUTPUT_DIRECTORY,
                              default_value=os.path.join(kamzik3.session.get_value(ATTR_LOG_DIRECTORY), "Scan_logs"),
                              readonly=True)
        self.create_attribute(ATTR_TEMPLATE_DIRECTORY,
                              default_value=os.path.join(kamzik3.session.get_value(ATTR_RESOURCE_DIRECTORY),
                                                         "Scan_templates"), readonly=True)
        self.create_attribute(ATTR_MAX_RUNNING_MACROS, default_value=1, default_type=np.uint16)
        self.create_attribute(ATTR_MAX_MACRO_HISTORY, default_value=1, default_type=np.uint16, min_value=1)

    def _init_templates(self):
        pass

    def command(self, command, callback=None, with_token=False, returning=True):
        raise DeviceError("Device does not accept any commands.")

    def set_macro_templates_directory(self, value=None):
        if value is not None:
            template_directory_path = os.path.join(kamzik3.session.get_value(ATTR_RESOURCE_DIRECTORY), value)
        else:
            template_directory_path = self.get_value(ATTR_TEMPLATE_DIRECTORY)

        if not os.path.exists(template_directory_path):
            self.logger.info(u"Template directory {} does not exists, trying to create it.".format(
                template_directory_path))
            os.makedirs(template_directory_path)

        self.set_value(ATTR_TEMPLATE_DIRECTORY, template_directory_path)
        self.read_scan_templates_directory()

    def read_scan_templates_directory(self):
        self.loaded_templates = {}
        dir_path = self.get_value(ATTR_TEMPLATE_DIRECTORY)
        files = [os.path.join(dir_path, f) for f in os.listdir(dir_path) if os.path.isfile(os.path.join(dir_path, f))]
        files.sort()
        loaded_templates_count = 0
        for file_item in files:
            with open(file_item, "r") as fp:
                template = yaml.load(fp, Loader=yaml.Loader)
                self.loaded_templates[template.get("title")] = (template, file_item)
                loaded_templates_count += 1
        self.set_attribute([ATTR_TEMPLATES_COUNT, VALUE], loaded_templates_count)

    @expose_method()
    def get_measurement_groups(self):
        return kamzik3.session.measurement_groups

    @expose_method()
    def get_templates(self):
        return self.loaded_templates

    @expose_method()
    def get_running_macros(self):
        return list(self.running_macros.keys())

    @expose_method()
    def get_finished_macros(self):
        return list(self.finished_macros.keys())

    @expose_method()
    def get_uploaded_macros(self):
        return list(self.uploaded_macros.keys())

    @expose_method()
    def get_running_macros_meta(self):
        macros_meta_list = []
        for running_macro in self.running_macros.values():
            macros_meta_list.append({
                "id": running_macro.device_id,
                "header": running_macro.macro_logger.get_value(ATTR_HEADER)
            })
        return macros_meta_list

    @expose_method()
    def broadcast_macro_header(self):
        """
        This is just a patch function to support Onda.
        TODO: remove this function after Onda is properly communicating with Kamzik3
        :return:
        """
        for running_macro in self.running_macros.values():
            attribute_object = running_macro.macro_logger[ATTR_HEADER]
            header = "{}.{}".format(attribute_object.attribute_id, VALUE)
            kamzik3.session.publisher.push_message(header, running_macro.macro_logger.get_value(ATTR_HEADER))

    @expose_method({"macro_id": "ID of uploaded macro"})
    def get_macro_template(self, macro_id):
        try:
            return self.loaded_templates[self.running_macros.get(macro_id).template_id]
        except (AttributeError, TypeError) as e:
            pass
        try:
            return self.loaded_templates[self.uploaded_macros.get(macro_id).template_id]
        except (AttributeError, TypeError) as e:
            pass
        try:
            return self.loaded_templates[self.finished_macros.get(macro_id).template_id]
        except (AttributeError, TypeError) as e:
            error_message = "Could not get Template: {}".format(e)
            self.logger.error(error_message)
            raise DeviceError(error_message)

    @expose_method({"template_metadata": "YAML serialized template metadata"})
    def create_template(self, template_metadata):
        try:
            template_data = oyaml.load(template_metadata, Loader=yaml.Loader)
            file_path = os.path.join(self.get_value(ATTR_TEMPLATE_DIRECTORY),
                                     template_data["title"].replace(" ", "_") + ".stpl")
        except (AttributeError, TypeError) as e:
            raise DeviceError("Template could not be created. {}".format(e))
        if file_path != "":
            with open(file_path, "w") as fp:
                fp.write(template_metadata)
                self.logger.info("New template {} saved.".format(file_path))
            self.read_scan_templates_directory()
            return file_path
        else:
            raise DeviceError("Cannot save template {}.".format(template_data["title"]))

    @expose_method({"template_id": "ID of template to be removed"})
    def remove_template(self, template_id):
        try:
            template, file_item = self.loaded_templates.get(template_id, None)
            os.remove(file_item)
        except (FileNotFoundError, OSError, TypeError) as e:
            error_message = "Template could not be removed: {}".format(e)
            self.logger.error(error_message)
            raise DeviceError(error_message)

        self.logger.info(u"Removing template {} at {}.".format(template_id, file_item))
        self.read_scan_templates_directory()

    @expose_method({"template_id": "Name of macro template", "macro_data": "YAML serialized macro data"})
    def upload_macro(self, template_id, macro_data):
        macro_id = "{}_{}".format(self.get_value(ATTR_MACRO_PREFIX), self.get_macro_count())
        macro_data = re.sub("^device_id:.*", "device_id: {}".format(macro_id), macro_data, flags=re.MULTILINE)
        macro = yaml.load(macro_data, Loader=oyaml.Loader)
        if macro is None:
            raise DeviceError("Macro could not be uploaded")
        try:
            macro.template_id = template_id
        except (AttributeError, TypeError) as e:
            raise DeviceError("Macro could not be uploaded. {}".format(e))

        return self.add_macro(macro)

    def add_macro(self, macro):
        self.uploaded_macros[macro.device_id] = macro
        self.set_value(ATTR_UPLOADED_COUNT, self.get_value(ATTR_UPLOADED_COUNT) + 1)
        self.set_macro_log(macro.device_id)

        if len(self.finished_macros) > self.get_value(ATTR_MAX_MACRO_HISTORY):
            self.finished_macros.popitem(last=False)[1].remove()
            self.set_value(ATTR_FINISHED_COUNT, len(self.finished_macros))
        cb = lambda value, macro=macro: self.macro_status_update(value, macro)
        macro.attach_attribute_callback(ATTR_STATUS, cb, key_filter=VALUE)
        return macro.common_id

    @expose_method({"macro_id": "ID of uploaded macro"})
    def start_macro(self, macro_id):
        if self.get_value(ATTR_RUNNING_COUNT) >= self.get_value(ATTR_MAX_RUNNING_MACROS):
            raise DeviceError("Maximum running macros count reached.")
        if macro_id not in self.uploaded_macros:
            raise DeviceError("Macro {} was not found on MacroServer".format(macro_id))
        macro = self.uploaded_macros.pop(macro_id)
        self.set_value(ATTR_UPLOADED_COUNT, len(self.uploaded_macros))
        macro.start()

    @expose_method()
    def clear_uploaded_macros(self):
        for macro in list(self.uploaded_macros.values()):
            del self.uploaded_macros[macro.common_id]
            macro.remove()
        self.set_value(ATTR_UPLOADED_COUNT, len(self.uploaded_macros))

    @expose_method()
    def clear_finished_macros(self):
        for macro in list(self.finished_macros.values()):
            del self.finished_macros[macro.common_id]
            macro.remove()
        self.set_value(ATTR_FINISHED_COUNT, len(self.finished_macros))

    def macro_status_update(self, value, macro):
        with self.update_lock:
            if value in (STATUS_IDLE, STATUS_ERROR):
                if macro.common_id in self.running_macros:
                    del self.running_macros[macro.common_id]
                self.finished_macros[macro.common_id] = macro
                self.set_value(ATTR_RUNNING_COUNT, len(self.running_macros))
                self.set_value(ATTR_FINISHED_COUNT, len(self.finished_macros))
                if value == STATUS_ERROR:
                    self.handle_command_error(macro.common_id,
                                              "{}: {}".format(macro.common_id, macro.get_value(ATTR_LAST_ERROR)))
            elif value == STATUS_BUSY:
                self.running_macros[macro.common_id] = macro
                self.set_value(ATTR_UPLOADED_COUNT, len(self.uploaded_macros))
                self.set_value(ATTR_RUNNING_COUNT, len(self.running_macros))

    def set_macro_log(self, macro_id):
        macro = self.uploaded_macros.get(macro_id)
        output_file, macro_index = macro.setup_output_file(self.get_value(ATTR_OUTPUT_DIRECTORY),
                                                           self.get_value(ATTR_MACRO_PREFIX))
        current_template, _ = self.loaded_templates.get(macro.template_id, ({}, None))
        header_lines = [
            u"Log version: {}".format("0.5"),
            u"Session ID: {}".format(kamzik3.session.device_id),
            u"Template: {}".format(self.get_value(ATTR_MACRO_PREFIX), macro.template_id),
            u"Prefix: {}".format(self.get_value(ATTR_MACRO_PREFIX)),
            u"Index: {}".format(macro_index),
            u"Comment: {}".format(macro.comment),
            u"Output file: {}".format(output_file),
            u"Points count: {}".format(macro.get_total_points_count()),
            u"Repeat count: {}".format(macro.repeat_count),
            u"Measurement groups: {}".format(current_template.get("measurement_groups", None)),
            u"-" * 32,
        ]
        header_lines += macro.get_output_header()
        # Check if attribute log is enabled for session
        if self.session.get_value(ATTR_ALLOW_ATTRIBUTE_LOG):
            attribute_logger = self.session.get_device("AttributeLogger")
            header = attribute_logger.get_value(ATTR_HEADER)[2:]
            sep = attribute_logger.separator
            log_line = attribute_logger.get_value(ATTR_LAST_LOG_LINE)
            now = datetime.now()
            # Add last logged attributes to the Scan header
            header_lines += [u"Session logged attributes", header,
                             sep.join([now.strftime("%Y-%m-%d %H:%M:%S"), str(time.time()), log_line]), u"-" * 32]
        # Set macro logger header
        macro.macro_logger.preset_header = u"\n".join(u"# {0}".format(s) for s in header_lines) + u"\n"

        # Get through all items in macro and set attributes to log for each macro step
        for index, chain_item in enumerate(macro.chain):
            if isinstance(chain_item, Scan):
                # It's Scan step
                log_attribute_name = chain_item.step_attributes.get("output", chain_item.step_attributes["attribute"])
                for measurement_group in current_template.get("measurement_groups", []):
                    for device, attribute in kamzik3.session.measurement_groups[measurement_group]:
                        macro.add_output_attribute(device, attribute)

                macro.add_output_attribute(chain_item.step_attributes["device_id"], log_attribute_name)
            elif isinstance(chain_item, StepGenerator):
                # It's Step generator
                for measurement_group in current_template.get("measurement_groups", []):
                    for device, attribute in kamzik3.session.measurement_groups[measurement_group]:
                        macro.add_output_attribute(device, attribute)

    def set_macro_prefix(self, value):
        """
        Set prefix for each new macro, which will be executed on macro-server.
        :param value:
        :return:
        """
        self.set_raw_value(ATTR_MACRO_PREFIX, value)
        self.set_macro_log_directory("{}_logs".format(value))
        self.set_macro_templates_directory("{}_templates".format(value))
        self.get_macro_count()

    @expose_method()
    def get_macro_count(self):
        """
        Macro output directory for highest postfix which is our macro count number.
        :return: int
        """
        macro_prefix = self.get_value(ATTR_MACRO_PREFIX)
        output_directory_path = self.get_value(ATTR_OUTPUT_DIRECTORY)
        next_index = get_next_file_index(output_directory_path, re.compile("{}_(\d+)\.log".format(macro_prefix)))
        self.set_attribute([ATTR_MACRO_COUNT, VALUE], next_index)
        return next_index

    def set_macro_log_directory(self, value=None):
        """
        Set macro base log directory.
        All macro log files will go into this directory.
        :param value:
        :return:
        """
        if value is not None:
            log_directory = kamzik3.session.get_value(ATTR_LOG_DIRECTORY)
            if log_directory is None:
                raise DeviceError(u"Cannot set macro log directory. Attribute 'Log directory' is not set.")
            log_directory = os.path.join(log_directory, value)
        else:
            log_directory = self.get_value(ATTR_OUTPUT_DIRECTORY)

        self.logger.info(u"Setting macro log directory to: {}".format(log_directory))
        if not os.path.exists(log_directory):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(log_directory))
            os.makedirs(log_directory)

        self.set_value(ATTR_OUTPUT_DIRECTORY, log_directory)



from PyQt5.QtCore import QAbstractTableModel, Qt, pyqtSignal, pyqtSlot, QModelIndex, QItemSelection
from PyQt5.QtGui import QShowEvent
from PyQt5.QtWidgets import QWidget

import kamzik3
from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.gui.attributeDeviceDisplayWidget import AttributeDeviceDisplayWidget
from kamzik3.gui.templates.attributeSetpointHistoryTemplate import Ui_Form
from kamzik3.snippets.snippetsUnits import seconds_to_datetime, device_units
from kamzik3.snippets.snippetsWidgets import show_question_dialog, show_error_message


class _view(QAbstractTableModel):

    def __init__(self, parent=None):
        self.model_data = AttributeDeviceDisplayWidget.setpoints_history
        self.header = ["Datetime", "Device", "Attribute", "Previous value", "New setpoint"]
        QAbstractTableModel.__init__(self, parent)

    def columnCount(self, parent=None, *args, **kwargs):
        return 5

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.model_data)

    def data(self, index, role=None):
        row, column = index.row(), index.column()

        if role == Qt.DisplayRole:
            if column == 0:
                return "{}".format(seconds_to_datetime(self.model_data[-row - 1][column]))
            else:
                return "{}".format(self.model_data[-row - 1][column])

    def headerData(self, p_int, Qt_Orientation, role=None):
        if role == Qt.DisplayRole and Qt_Orientation == Qt.Horizontal:
            return self.header[p_int]
        elif role == Qt.DisplayRole and Qt_Orientation == Qt.Vertical:
            return p_int + 1

    def flags(self, index):
        if index.column() <= 2:
            return Qt.ItemIsEnabled
        else:
            return Qt.ItemIsSelectable | Qt.ItemIsEnabled


class AttributeSetpointHistoryWidget(Ui_Form, QWidget):
    sigErrorCheck = pyqtSignal("PyQt_PyObject", "PyQt_PyObject")

    def __init__(self, config=None, parent=None):
        QWidget.__init__(self, parent)
        self.config = config
        if config is None:
            self.config = {}
        self.setupUi(self)
        self.table_history_view.doubleClicked.connect(self.on_double_click)
        self.sigErrorCheck.connect(self.slot_check_error)
        self.table_history_view.setModel(_view())
        self.table_history_view.selectionModel().selectionChanged.connect(self.current_selection_changed)
        self.table_history_view.horizontalHeader().resizeSection(0, 120)
        self.table_history_view.horizontalHeader().resizeSection(3, 200)
        self.button_set_value.setDisabled(True)

    def yaml_mapping(self):
        mapping = super().yaml_mapping()
        del mapping["parent"]
        return mapping

    def set_attribute_value(self):
        try:
            index = self.table_history_view.selectedIndexes()[0]
            row, column = -index.row() - 1, index.column()
            if index.column() >= 3:
                model_data = self.table_history_view.model().model_data
                (ts, device_id, attribute, previous_value, new_setpoint) = model_data[row]
                question = show_question_dialog(
                    u"Do You really want to set {} {} to {} ?".format(device_id, attribute, model_data[row][column]),
                    parent=self)
                if question:
                    device = kamzik3.session.get_device(device_id)
                    value_in_device_units = device_units(device, attribute, model_data[row][column])
                    device.set_attribute(attribute + [VALUE], value_in_device_units.m, callback=self.sigErrorCheck.emit)
        except (KeyError, DeviceError) as e:
            show_error_message(e, parent=self)
        except IndexError:
            show_error_message(u"Invalid setpoint selection", parent=self)

    @pyqtSlot("PyQt_PyObject", "PyQt_PyObject")
    def slot_check_error(self, attribute, output):
        if attribute == RESPONSE_ERROR:
            show_error_message(output, parent=self)

    @pyqtSlot(QModelIndex)
    def on_double_click(self, index):
        if index.column() >= 3:
            self.set_attribute_value()

    @pyqtSlot(QItemSelection, QItemSelection)
    def current_selection_changed(self, selected, deselected):
        if len(selected) > 0:
            if selected[0].topLeft().column() >= 3:
                self.button_set_value.setEnabled(True)
                return

        self.button_set_value.setEnabled(False)

    def reset(self):
        if show_question_dialog(u"Do You want to reset setpoint history?", title=u"Reset setpoint history",
                                parent=self):
            AttributeDeviceDisplayWidget.setpoints_history.clear()
            self.refresh()

    @pyqtSlot()
    def refresh(self):
        self.table_history_view.model().layoutChanged.emit()

    def showEvent(self, a0: QShowEvent) -> None:
        self.refresh()
        super().showEvent(a0)



import time

from PyQt5.QtCore import QSize
from PyQt5.QtWidgets import QSizePolicy
from pyqtgraph.widgets.RawImageWidget import RawImageWidget

from kamzik3.constants import *
from kamzik3.gui.attributeWidgets.attributeWidget import AttributeWidget
from kamzik3.gui.fastDisplay.imageView import ImageView


class AttributeImageWidget(AttributeWidget):
    lt = time.perf_counter_ns()

    def _set_input_widget(self):
        self.input_widget = ImageView()
        if self.attribute[DESCRIPTION] is not None:
            self.input_widget.setToolTip(self.attribute[DESCRIPTION])
        self.input_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def set_value(self, value):
        self.input_widget.slot_new_frame(value)

    def get_attribute_value(self):
        return None

    def get_widget_value(self):
        return None



from PyQt5.QtWidgets import QSizePolicy

from kamzik3.constants import *
from kamzik3.gui.attributeWidgets.attributeWidget import AttributeWidget
from kamzik3.snippets.snippetsWidgets import CustomTextEdit


class AttributeTextWidget(AttributeWidget):

    def _set_input_widget(self):
        self.input_widget = CustomTextEdit()
        if self.attribute[READONLY]:
            self.input_widget.setReadOnly(True)
        else:
            self.input_widget.setStyleSheet(".CustomLineEdit {background:#e6ffd9}")
            if self.attribute[DESCRIPTION] is not None:
                self.input_widget.setToolTip(self.attribute[DESCRIPTION])

        self.input_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

    def set_value(self, value):
        self.input_widget.setPlainText(str(value))

    def get_attribute_value(self):
        return self.input_widget.text()

    def get_widget_value(self):
        return self.input_widget.text()



import os
import re
from threading import Event

from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetDataAccess import get_next_file_index
from kamzik3.snippets.snippetsDecorators import expose_method


class DeviceRecorder(Device):

    def __init__(self, device, attribute, device_id=None, config=None):
        self.stopped = Event()
        self.frame_delay = 0
        Device.__init__(self, device_id, config)
        self.connect()

    def _init_attributes(self):
        Device._init_attributes(self)
        self.create_attribute(ATTR_OUTPUT_DIRECTORY, default_value="./")
        self.create_attribute(ATTR_FourCC_CODEC, default_value="X264",
                              default_type=["X264", "XVID", "MJPG"])
        self.create_attribute(ATTR_RECORDING, default_value=False, default_type=bool, readonly=True)
        self.create_attribute(ATTR_FILENAME_PREFIX, default_value="video")
        self.create_attribute(ATTR_OUTPUT_FILENAME, default_value=None, readonly=True)

    def setup_output_file(self):
        output_directory_path = self.get_value(["Record", ATTR_OUTPUT_DIRECTORY])
        file_prefix = self.get_value(["Record", ATTR_FILENAME_PREFIX])
        if not os.path.exists(output_directory_path):
            self.logger.info(u"Output directory {} does not exists, trying to create it.".format(output_directory_path))
            os.makedirs(output_directory_path)
        next_index = get_next_file_index(output_directory_path, re.compile("{}_(\d+)\.mp4".format(file_prefix)))
        output_filename = os.path.join(output_directory_path, "{}_{}.mp4".format(file_prefix, next_index))
        self.set_value(["Record", ATTR_OUTPUT_FILENAME], output_filename)

    @expose_method()
    def start_recording(self):
        self.setup_output_file()

    @expose_method()
    def stop_recording(self):
        pass



import time
from _ctypes import byref
from ctypes import create_string_buffer, c_uint32, CDLL, c_uint16, c_int, c_int8, c_uint8, c_int64

import bidict
import numpy as np
from pint import UndefinedUnitError

from kamzik3.snippets.snippetsUnits import device_units

from kamzik3 import DeviceError, DeviceUnitError
from kamzik3.constants import *
from kamzik3.devices.deviceChannel import DeviceChannel
from kamzik3.devices.deviceOsLibrary import DeviceOsLibrary
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsTimer import PreciseCallbackTimer

MODE_PROFILE_POSITION = 1
MODE_PROFILE_VELOCITY = 3
MODE_HOMING = 6
MODE_POSITION = 8
MODE_VELOCITY = 9
MODE_CURRENT = 10

GROUP_VELOCITY_PROFILE = "Velocity profile"
GROUP_POSITION_PROFILE = "Position profile"


class DeviceEpos(DeviceOsLibrary):

    def __init__(self, library, device_name, protocol, interface, port, device_id=None, config=None):
        self.device_name = create_string_buffer(device_name.encode())
        self.protocol = create_string_buffer(protocol.encode())
        self.interface = create_string_buffer(interface.encode())
        self.port = create_string_buffer(port.encode())
        self.handle = 0
        self.pError = c_uint32()
        DeviceOsLibrary.__init__(self, library, device_id, config)

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_IDLE)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.connected = True
        _finish_configuration()

    def connect(self, *args):
        """
        Call only this function to connect devices to port / socket / library / ...
        :param args: connect attributes
        """
        self.connecting = True
        self.handle_configuration_event()
        try:
            self.lib = CDLL(self.os_library)
            self.handle = self.check_error(
                self.lib.VCS_OpenDevice(byref(self.device_name), byref(self.protocol), byref(self.interface),
                                        byref(self.port), byref(self.pError)))
        except OSError as e:
            raise DeviceError("Error loading library {}. {}".format(self.os_library, e))
        self.connected = True
        self.connecting = False
        self.set_status(STATUS_IDLE)

    def check_error(self, error_code, raise_exception=True):
        """
        Check error response for executed command.
        :param error_code: error code response
        :param raise_exception: raise exception if True
        :return:
        """
        if error_code == 0:
            error_message, error_code = self.get_error()
            if raise_exception and error_code != 0:
                raise DeviceError("Error: %s" % error_message)
        return error_code

    def get_error(self, pError=None):
        """
        Get error description of given code.
        :param pError: error code
        :return:
        """
        if pError is None:
            pError = self.pError
        pErrorInfo = create_string_buffer(255)
        if self.lib.VCS_GetErrorInfo(pError, byref(pErrorInfo), 255) != 0:
            return "Code: %s, %s" % (pError.value, pErrorInfo.value), pError.value
        else:
            return "Code: %s, %s" % (pError.value, "Unknown error"), pError.value

    def disconnect(self):
        self.lib.VCS_CloseDevice(self.handle, byref(self.pError))
        return super().disconnect()


class DeviceEposLinux(DeviceEpos):

    def __init__(self, device_name, protocol, interface, port, device_id=None, config=None):
        DeviceEpos.__init__(self, "libEposCmd.so", device_name, protocol, interface, port, device_id, config)


class DeviceEposChannel(DeviceChannel):
    polling = False
    operation_modes = bidict.frozenbidict({
        MODE_PROFILE_POSITION: u"Profile position",
        MODE_PROFILE_VELOCITY: u"Profile velocity",
        MODE_HOMING: u"Homing",
        MODE_POSITION: u"Position",
        MODE_VELOCITY: u"Velocity",
        MODE_CURRENT: u"Current",
    })
    velocity_units = bidict.frozenbidict({
        u"Standard": 0,
        u"Deci": -1,
        u"Centi": -2,
        u"Milli": -3,
    })

    def _init_attributes(self):
        DeviceChannel._init_attributes(self)

        self.create_attribute(ATTR_HARDWARE_VERSION, default_value=None, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_SOFTWARE_VERSION, default_value=None, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_APPLICATION_NUMBER, default_value=None, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_APPLICATION_VERSION, default_value=None, readonly=True, default_type=np.uint32)
        self.create_attribute(ATTR_AUTO_STOP, default_value=True, readonly=False, default_type=bool)
        self.create_attribute(ATTR_ON_TARGET, default_value=True, readonly=True, default_type=bool)
        self.create_attribute(ATTR_REVOLUTIONS, readonly=True, default_type=np.int64)
        self.create_attribute(ATTR_OPERATION_MODE, default_value=None, readonly=True,
                              default_type=self.operation_modes.inverse.keys(), set_function=self.set_operation_mode)
        self.create_attribute(ATTR_MOTOR_EXCITATION, default_value=False, default_type=bool,
                              set_function=self.set_motor_enabled)
        self.create_attribute(ATTR_VELOCITY_UNITS, default_value=None, readonly=True,
                              default_type=self.velocity_units.keys(), set_function=self.set_velocity_units)
        self.create_attribute(ATTR_VELOCITY, default_value=None, readonly=False, unit="rpm",
                              default_type=np.float, set_function=self.set_velocity)
        self.create_attribute(ATTR_ABSOLUTE_POSITION, default_value=None, readonly=False, unit="deg",
                              default_type=np.float, set_function=self.set_absolute_position)
        self.create_attribute(ATTR_POSITION, default_value=None, readonly=False, unit="deg",
                              default_type=np.float, set_function=self.set_position)
        self.create_attribute(ATTR_ACCELERATION, default_value=None, readonly=False, group=GROUP_VELOCITY_PROFILE,
                              default_type=np.float, set_function=lambda v: self.set_velocity_profile(acceleration=v))
        self.create_attribute(ATTR_DECELERATION, default_value=None, readonly=False, group=GROUP_VELOCITY_PROFILE,
                              default_type=np.float, set_function=lambda v: self.set_velocity_profile(deceleration=v))
        self.create_attribute(ATTR_ACCELERATION, default_value=None, readonly=False, group=GROUP_POSITION_PROFILE,
                              default_type=np.float, set_function=lambda v: self.set_position_profile(acceleration=v))
        self.create_attribute(ATTR_DECELERATION, default_value=None, readonly=False, group=GROUP_POSITION_PROFILE,
                              default_type=np.float, set_function=lambda v: self.set_position_profile(deceleration=v))
        self.create_attribute(ATTR_VELOCITY, default_value=None, readonly=False, group=GROUP_POSITION_PROFILE,
                              default_type=np.float, unit="rpm", decimals=4, set_function=lambda v: self.set_position_profile(velocity=v))

    def handle_configuration(self):
        if self.configured:
            return

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.configured = True
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.get_device_firmware_version()
        self.get_operation_mode()
        self.get_enabled_state()
        self.get_velocity_units()
        self.get_velocity_profile()
        self.get_position_profile()
        _finish_configuration()

    def get_device_firmware_version(self):
        '''
        VCS_GetVersion(HANDLE
        KeyHandle, WORD
        NodeId, WORD * pHardwareVersion, WORD *
        pSoftwareVersion, WORD * pApplicationNumber, WORD * pApplicationVersion, DWORD * pErrorCode)
        '''
        hw_version = c_uint16()
        sw_version = c_uint16()
        app_number = c_uint16()
        app_version = c_uint32()
        self.device.check_error(
            self.device.lib.VCS_GetVersion(self.device.handle, self.channel, byref(hw_version), byref(sw_version),
                                           byref(app_number),
                                           byref(app_version), byref(self.device.pError)))
        self.set_value(ATTR_HARDWARE_VERSION, hw_version.value)
        self.set_value(ATTR_SOFTWARE_VERSION, sw_version.value)
        self.set_value(ATTR_APPLICATION_NUMBER, app_number.value)
        self.set_value(ATTR_APPLICATION_VERSION, app_version.value)
        self.set_value([ATTR_APPLICATION_VERSION], app_version.value)

    def get_movement_state(self):
        state = c_int()
        self.device.check_error(
            self.device.lib.VCS_GetMovementState(self.device.handle, self.channel, byref(state),
                                                 byref(self.device.pError))
        )
        self.set_value(ATTR_ON_TARGET, bool(state.value))
        return state.value

    def get_state(self):
        state = c_uint32()
        self.device.check_error(
            self.device.lib.VCS_GetState(self.device.handle, self.channel, byref(state), byref(self.device.pError))
        )
        self.set_status(STATUS_BUSY if state.value else STATUS_IDLE)
        return state.value

    def get_enabled_state(self):
        state = c_int8()
        self.device.check_error(
            self.device.lib.VCS_GetEnableState(self.device.handle, self.channel, byref(state),
                                               byref(self.device.pError))
        )
        self.set_raw_value(ATTR_MOTOR_EXCITATION, state.value)
        return state.value

    def get_operation_mode(self):
        operation_mode = c_uint32()
        self.device.check_error(
            self.device.lib.VCS_GetOperationMode(self.device.handle, self.channel, byref(operation_mode),
                                                 byref(self.device.pError))
        )
        self.set_raw_value(ATTR_OPERATION_MODE, self.operation_modes.get(operation_mode.value, None))
        return operation_mode.value

    def get_velocity_units(self):
        dimension = c_uint8()
        notation = c_int8()
        self.device.check_error(
            self.device.lib.VCS_GetVelocityUnits(self.device.handle, self.channel, byref(dimension), byref(notation),
                                                 byref(self.device.pError))
        )
        self.set_raw_value(ATTR_VELOCITY_UNITS, self.velocity_units.inverse.get(notation.value))
        return self.velocity_units.inverse.get(notation.value)

    def get_velocity(self):
        velocity = c_int64()
        self.device.check_error(
            self.device.lib.VCS_GetVelocityIsAveraged(self.device.handle, self.channel, byref(velocity),
                                                      byref(self.device.pError))
        )
        self.set_raw_value(ATTR_VELOCITY, velocity.value)
        return velocity.value

    def get_velocity_profile(self):
        acceleration = c_int()
        deceleration = c_int()
        self.device.check_error(
            self.device.lib.VCS_GetVelocityProfile(self.device.handle, self.channel, byref(acceleration),
                                                   byref(deceleration), byref(self.device.pError))
        )
        self.set_raw_value([GROUP_VELOCITY_PROFILE, ATTR_ACCELERATION], acceleration.value)
        self.set_raw_value([GROUP_VELOCITY_PROFILE, ATTR_DECELERATION], deceleration.value)
        return acceleration.value, deceleration.value

    def get_position(self):
        position = c_int64()
        self.device.check_error(
            self.device.lib.VCS_GetPositionIs(self.device.handle, self.channel, byref(position),
                                              byref(self.device.pError))
        )
        attribute = self.get_attribute(ATTR_ABSOLUTE_POSITION)
        self.set_raw_value(ATTR_ABSOLUTE_POSITION, position.value)
        self.set_raw_value(ATTR_POSITION, attribute.apply_offset_factor(position.value) % 360)
        self.set_raw_value(ATTR_REVOLUTIONS, int(attribute.apply_offset_factor(position.value) / 360))
        return position.value

    def get_position_profile(self):
        velocity = c_int()
        acceleration = c_int()
        deceleration = c_int()
        self.device.check_error(
            self.device.lib.VCS_GetPositionProfile(self.device.handle, self.channel, byref(velocity),
                                                   byref(acceleration), byref(deceleration), byref(self.device.pError))
        )
        self.set_raw_value([GROUP_POSITION_PROFILE, ATTR_VELOCITY], velocity.value)
        self.set_raw_value([GROUP_POSITION_PROFILE, ATTR_ACCELERATION], acceleration.value)
        self.set_raw_value([GROUP_POSITION_PROFILE, ATTR_DECELERATION], deceleration.value)
        return velocity.value, acceleration.value, deceleration.value

    def set_operation_mode(self, value):
        operation_mode = self.operation_modes.inverse[value]
        self.device.check_error(
            self.device.lib.VCS_SetOperationMode(self.device.handle, self.channel, c_int8(operation_mode),
                                                 byref(self.device.pError))
        )
        self.get_operation_mode()

    def set_velocity(self, velocity):
        velocity = int(velocity)
        self.logger.debug("Set velocity to: {} RPM".format(velocity))
        self.set_motor_enabled(True)
        self.set_operation_mode(self.operation_modes[MODE_PROFILE_VELOCITY])
        self.device.check_error(
            self.device.lib.VCS_MoveWithVelocity(self.device.handle, self.channel, c_int64(velocity),
                                                 byref(self.device.pError))
        )

    def set_absolute_position(self, position, absolute=True):
        position = int(position)
        self.logger.debug("Set absolute position to: {} deg".format(position))
        self.set_motor_enabled(True)
        self.set_operation_mode(self.operation_modes[MODE_PROFILE_POSITION])
        self.device.check_error(
            self.device.lib.VCS_MoveToPosition(self.device.handle, self.channel, c_int64(position), c_int(absolute),
                                               c_int(True), byref(self.device.pError))
        )

    def set_position(self, position):
        position = float(position)
        self.logger.debug("Set position to: {} deg".format(position))
        attribute = self.get_attribute(ATTR_ABSOLUTE_POSITION)
        self.set_absolute_position(attribute.remove_offset_factor((self[ATTR_REVOLUTIONS].value().m * 360) + position))

    def set_motor_enabled(self, value):
        '''
        Enable motor for remote control
        @param value: enable flag (0 - Off, 1 - On)
        @type value: int
        '''
        if value:
            self.logger.info("Motor enabled")
            self.device.check_error(
                self.device.lib.VCS_SetEnableState(self.device.handle, self.channel, byref(self.device.pError)))
        else:
            self.logger.info("Motor disabled")
            self.device.check_error(
                self.device.lib.VCS_SetDisableState(self.device.handle, self.channel, byref(self.device.pError)))

    def set_velocity_units(self, value):
        self.set_motor_enabled(False)
        dimension = c_uint8()
        notation = c_int8()
        self.device.check_error(
            self.device.lib.VCS_GetVelocityUnits(self.device.handle, self.channel, byref(dimension), byref(notation),
                                                 byref(self.device.pError))
        )
        notation = c_int8(self.velocity_units.get(value, 0))
        self.device.check_error(
            self.device.lib.VCS_SetVelocityUnits(self.device.handle, self.channel, dimension, notation,
                                                 byref(self.device.pError))
        )

    def set_velocity_profile(self, acceleration=None, deceleration=None):
        acceleration = c_int(int(
            self.get_value([GROUP_VELOCITY_PROFILE, ATTR_ACCELERATION]) if acceleration is None else acceleration))
        deceleration = c_int(int(
            self.get_value([GROUP_VELOCITY_PROFILE, ATTR_DECELERATION]) if deceleration is None else deceleration))
        self.device.check_error(
            self.device.lib.VCS_SetVelocityProfile(self.device.handle, self.channel,
                                                   acceleration, deceleration, byref(self.device.pError))
        )

    def set_position_profile(self, velocity=None, acceleration=None, deceleration=None):
        velocity = c_int(int(
            self.get_value([GROUP_VELOCITY_PROFILE, ATTR_VELOCITY]) if velocity is None else velocity))
        acceleration = c_int(int(
            self.get_value([GROUP_VELOCITY_PROFILE, ATTR_ACCELERATION]) if acceleration is None else acceleration))
        deceleration = c_int(int(
            self.get_value([GROUP_VELOCITY_PROFILE, ATTR_DECELERATION]) if deceleration is None else deceleration))
        self.device.check_error(
            self.device.lib.VCS_SetPositionProfile(self.device.handle, self.channel, velocity,
                                                   acceleration, deceleration, byref(self.device.pError))
        )

    @expose_method()
    def stop(self):
        self.device.check_error(
            self.device.lib.VCS_HaltVelocityMovement(self.device.handle, self.channel, byref(self.device.pError))
        )
        self.set_motor_enabled(False)

    def poller_interval(self):
        self.get_velocity()
        self.get_position()
        self.get_enabled_state()
        self.get_state()
        if self.get_value(ATTR_AUTO_STOP) and self.get_value(ATTR_OPERATION_MODE) == self.operation_modes[
            MODE_PROFILE_POSITION] and self.get_movement_state():
                self.stop()
        elif self.get_value(ATTR_AUTO_STOP) and self.get_value(ATTR_OPERATION_MODE) == self.operation_modes[
            MODE_PROFILE_VELOCITY] and self.get_movement_state() and self.get_value(ATTR_VELOCITY) == 0 and self.get_attribute([ATTR_VELOCITY, SETPOINT]) == 0:
            self.stop()

    def start_polling(self):
        self.polling = True
        self.polling_timer = PreciseCallbackTimer(300, self.poller_interval)
        self.polling_timer.start()

    def stop_polling(self):
        self.polling = False
        self.polling_timer.stop()

    @expose_method({"position": ATTR_POSITION})
    def move_absolute(self, position):
        try:
            position = device_units(self, ATTR_POSITION, position)
            self.logger.info(u"Move motor channel to absolute position {}".format(position))
            return self.set_value(ATTR_POSITION, position.m)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    @expose_method({"position": ATTR_POSITION})
    def move_relative(self, position):
        try:
            position = device_units(self, ATTR_POSITION, position)
            self.logger.info(u"Move by relative position {}".format(position))
            return self.set_value(ATTR_POSITION, self[ATTR_POSITION].value() + position)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    @expose_method()
    def reset_position(self):
        self.set_operation_mode(self.operation_modes[MODE_HOMING])
        self.device.check_error(
            self.device.lib.VCS_FindHome(self.device.handle, self.channel, c_uint8(35),
                                                 byref(self.device.pError))
        )



import time

import cv2

from kamzik3.devices.general.deviceGrabber import DeviceGrabber

try:
    import mss
except ModuleNotFoundError:
    raise Exception("Module mss was not found.")
import numpy as np

from kamzik3.constants import *


class DeviceScreenGrabber(DeviceGrabber):

    def __init__(self, device_id=None, config=None):
        self.region = {"top": 0, "left": 0, "width": 640, "height": 480}
        DeviceGrabber.__init__(self, device_id, config)

    def _init_attributes(self):
        DeviceGrabber._init_attributes(self)
        self.create_attribute(ATTR_CAPTURE_REGION, default_value="0, 0, 640, 480", unit="px",
                              set_function=self.set_region)

    def set_region(self, value):
        top, left, width, height = value.split(",")
        # Part of the screen to capture
        self.region = {"top": int(top), "left": int(left), "width": int(width), "height": int(height)}

    def _acquire_loop(self):
        self.stopped.clear()
        self.frame_delay = 1. / self.get_value(ATTR_FRAMERATE)
        self.set_status(STATUS_BUSY)
        time_reference = time.perf_counter_ns()
        with mss.mss() as sct:
            while not self.stopped.wait(self.frame_delay - (time.perf_counter_ns() - time_reference) * 1e-9):
                img = np.array(sct.grab(self.region))
                self.set_value(["Output", "Frame"], cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
                time_reference += self.frame_delay * 1e9



from math import ceil

import numpy as np

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.deviceNiCard import DeviceNiCard, GROUP_DO

try:
    from PyDAQmx import *
except ImportError:
    raise DeviceError("PyDAQmx import error!")
from ctypes import *


class DeviceRheodyneBox(DeviceNiCard):
    max_channels_per_switch = 6

    def __init__(self, card_address, active_switches, device_id=None, config=None):
        self.active_switches = active_switches
        super().__init__(card_address, device_id, config)

    def _init_attributes(self):
        super()._init_attributes()
        self.create_attribute(ATTR_ACTIVE_SWITCHES, default_value=self.active_switches, readonly=True,
                              default_type=np.uint32, min_value=1)
        for i in range(self.active_switches):
            switch_name = "Switch{}".format(i + 1)
            self.attributes[switch_name] = {}
            sample_names = []
            for j in range(self.max_channels_per_switch):
                sample_name = "Sample {} name".format(j + 1)
                sample_names.append("Sample {}".format(j + 1))
                self.create_attribute(sample_name, switch_name, default_value="Sample {}".format(j + 1),
                                      set_function=lambda sn, i=i + 1, j=j + 1: self.set_sample_name(i, j, sn))

            self.create_attribute(ATTR_CURRENT_SAMPLE, switch_name, default_type=sample_names,
                                  set_function=lambda sn, i=i + 1: self.set_current_sample(i, sn))
            self.create_attribute(ATTR_CURRENT_CHANNEL, switch_name, default_type=np.uint16, min_value=1,
                                  max_value=self.max_channels_per_switch,
                                  set_function=lambda ch, i=i + 1: self.set_current_channel(i, ch))

    def set_current_sample(self, switch, sample_name):
        switch_name = "Switch{}".format(switch)
        self.logger.info("Setting Sample: {} on Switch: {}".format(sample_name, switch_name))
        default_type = self.attributes[switch_name][ATTR_CURRENT_SAMPLE][TYPE_LIST_VALUES]
        channel = default_type.index(sample_name) + 1
        self.set_current_channel(switch, channel)

    def set_current_channel(self, switch, channel):
        switch_name = "Switch{}".format(switch)
        channel = int(channel)
        self.logger.info("Setting Channel: {} on Switch: {}".format(channel, switch_name))
        default_type = self.attributes[switch_name][ATTR_CURRENT_SAMPLE][TYPE_LIST_VALUES]
        self.set_value([switch_name, ATTR_CURRENT_SAMPLE], default_type[channel - 1])
        self.set_value([switch_name, ATTR_CURRENT_CHANNEL], channel)
        port = ceil(switch / 2) - 1
        line_bits, line = ([16, 32, 64, 128], [1, 2, 4, 8])[switch % 2], ([4, 5, 6, 7], [0, 1, 2, 3])[switch % 2]

        try:
            self.write_do_task = self.task("set_digital_output_task")
            self.add_active_task(self.write_do_task, "set_digital_output_task")
            self.write_do_task.CreateDOChan("{}/port{}/line{}:{}".format(self.card_address, port, line[0], line[-1]),
                                            "", DAQmx_Val_ChanForAllLines)
            self.start_task(self.write_do_task.taskHandle.value)
            data = np.array([bool(channel & 1), bool(channel & 2), bool(channel & 4), bool(channel & 8)], dtype=uInt8)
            DAQmxWriteDigitalLines(self.write_do_task.taskHandle.value, 1, 1, 10.0, DAQmx_Val_GroupByChannel, data,
                                   None, None)
            self.stop_task("set_digital_output_task")
        except DAQError as e:
            raise DeviceError(u"Error setting digital output")

    def set_sample_name(self, switch, channel, name):
        switch_name = "Switch{}".format(switch)
        channel = int(channel)
        default_type = self.attributes[switch_name][ATTR_CURRENT_SAMPLE][TYPE_LIST_VALUES]
        default_type[channel - 1] = name
        self.set_attribute([switch_name, ATTR_CURRENT_SAMPLE, TYPE_LIST_VALUES], default_type)

    def handle_configuration(self):
        super().handle_configuration()
        self.read_digital_output()

        if self.config.get("macros"):
            for macro_name, macro_config in self.config.get("macros").items():
                setattr(self, macro_name,
                        lambda macro_config=macro_config: self.execute_configured_macro(macro_config))
                self.exposed_methods.append((macro_name, None))

    def execute_configured_macro(self, macro_config):
        switches, channels = macro_config
        if isinstance(channels, int):
            channels = [channels] * len(switches)
        if len(switches) != len(channels):
            raise DeviceError("Number of switches is not equal numbers of channels")
        for i, switch in enumerate(switches):
            channel = channels[i]
            self.set_current_channel(switch, channel)
        self.read_digital_output()

    def read_digital_output(self):
        try:
            self.read_do_task = self.task("get_digital_input_task")
            self.add_active_task(self.read_do_task, "get_digital_input_task")
            self.read_do_task.CreateDOChan("{}/port0/line0:7".format(self.card_address), "", DAQmx_Val_ChanForAllLines)
            self.read_do_task.CreateDOChan("{}/port1/line0:7".format(self.card_address), "",
                                           DAQmx_Val_ChanForAllLines)
            self.read_do_task.CreateDOChan("{}/port2/line0:7".format(self.card_address), "",
                                           DAQmx_Val_ChanForAllLines)
            self.start_task(self.read_do_task.taskHandle.value)
            task_buffer = np.zeros([1000], dtype=np.uint8)
            read = int32()
            bytesPerSamp = int32()
            DAQmxReadDigitalLines(self.read_do_task.taskHandle.value, 1, 10.0, DAQmx_Val_GroupByChannel, task_buffer,
                                  100, byref(read), byref(bytesPerSamp), None)

            word = ""
            switch = 1
            for i, do_line in enumerate(self[GROUP_DO].keys()):
                self.set_value([GROUP_DO, do_line], int(task_buffer[i]))
                if (i + 1) % 4 == 0 and switch <= self.active_switches:
                    switch_name = "Switch{}".format(switch)
                    default_type = self.attributes[switch_name][ATTR_CURRENT_SAMPLE][TYPE_LIST_VALUES]
                    channel = int(word, 2)
                    if channel > self.max_channels_per_switch:
                        channel = 1
                    self.set_value([switch_name, ATTR_CURRENT_SAMPLE], default_type[channel - 1])
                    self.set_value([switch_name, ATTR_CURRENT_CHANNEL], channel)
                    word = ""
                    switch += 1
                else:
                    word = str(task_buffer[i]) + word

            self.stop_task("get_digital_input_task")
        except DAQError:
            raise DeviceError(u"Error getting digital inputs")



import os
import re

from PyQt5.QtCore import Qt, pyqtSlot, QItemSelection, QItemSelectionModel, QDateTime
from PyQt5.QtWidgets import QWidget, QMdiSubWindow, QFileDialog
from pyqtgraph import exporters
from reportlab.platypus import SimpleDocTemplate, Image, LayoutError

import kamzik3
from kamzik3.constants import ATTR_CACHE_DIRECTORY
from kamzik3.gui.deviceWidget import DeviceWidget
from kamzik3.gui.logReader.logPlotView import LogPlotView, LogPlotViewMaster, QMdiSubWindowPlot
from kamzik3.gui.logReader.logViewHeaderModel import LogViewHeaderModel, LogHeaderSelectionModel
from kamzik3.gui.templates.logViewTemplate import Ui_Form
from kamzik3.snippets.snippetsUnits import get_scaled_time_duration
from kamzik3.snippets.snippetsWidgets import show_error_message

# View modes constants declaration

MODE_LIST = 0
MODE_ZOOM = 1


class LogView(Ui_Form, DeviceWidget):
    """
    Log Viewer widget
    Display plots in different modes.
    Use DeviceLogParser as a data source.
    """
    view_mode = MODE_LIST
    square_function_values = [x ** 2 for x in range(255)]
    pandas_dt = None
    master_window = None
    last_zoom_region = None
    view_tiled = True
    sub_windows_hint = (Qt.WindowSystemMenuHint | Qt.WindowCloseButtonHint | Qt.WindowMaximizeButtonHint)
    master_window_hint = (Qt.CustomizeWindowHint | Qt.WindowTitleHint)

    def setupUi(self, form):
        Ui_Form.setupUi(self, form)
        # Setup scrollbars width and height
        self.mdi_area.verticalScrollBar().setMaximumWidth(15)
        self.mdi_area.horizontalScrollBar().setMaximumHeight(15)
        # Connect activated sub-window to self.slot_window_activated method
        self.mdi_area.subWindowActivated.connect(self.slot_window_activated)

    def open(self, path, append=False, refresh=True, filter=None):
        """
        Open logfile.
        should be replaced.
        :param path:
        :return:
        """
        try:
            self.logger.info("Opening {}, append: {}, refresh: {}, filter: {}".format(path, append, refresh, filter))
            padas_dt = self.device.parse(path=path, format="mag_log", filter=filter)
            if append:
                self.pandas_dt = self.pandas_dt.append(padas_dt)
            else:
                self.pandas_dt = padas_dt
            if refresh:
                self.refresh()
        except Exception as e:
            error_message = "Error parsing file {}, probably unsupported format. {}".format(path, e)
            self.logger.error(error_message)
            show_error_message(error_message, self)
            return

    def refresh(self):
        self.slot_reset_all()
        filtered_columns = []
        for column, dtype in zip(self.pandas_dt.columns, self.pandas_dt.dtypes):
            if dtype == "bool":
                self.pandas_dt[column] = self.pandas_dt[column].astype(int, copy=False)
            if self.pandas_dt[column].isnull().all():
                # Ignore column if all values are NaN
                continue
            if dtype != "object":
                filtered_columns.append(column)
        model = LogViewHeaderModel(filtered_columns)
        self.input_header_items.setModel(model)
        max_selected_rows = self.input_max_selected_rows.value()
        selection_model = LogHeaderSelectionModel(model, max_selected_rows=max_selected_rows)
        self.input_header_items.setSelectionModel(selection_model)
        self.input_master_row.addItems(filtered_columns)

    def create_sub_window(self, id, master=False):
        """
        Create new sub-window with plot view widget.
        :param title: str
        :param master: bool
        :return:
        """
        window = QMdiSubWindowPlot(id)
        window.setWindowIcon(self.windowIcon())
        win_title, unit = id.split(" (")
        unit = unit[:-1]
        if not master:
            widget = LogPlotView(unit=unit, bottom_label=None)
            window.setWindowFlags(self.sub_windows_hint)
            window.closeEvent = lambda event, w=window: self.sub_window_closed(event, w)
            window.setMinimumSize = lambda w=window: self.set_master(w)
        else:
            widget = LogPlotViewMaster(unit=unit, bottom_label=None, color=(0, 0, 0))
            window.setWindowFlags(self.master_window_hint)
            self.master_window = window

            def region_chane_finished():
                self.master_zoom_changed(widget.lookingGlass.getRegion())

            widget.lookingGlass.sigRegionChangeFinished.connect(region_chane_finished)
            win_title = "{} - {}".format("Master", win_title)

        window.setWindowTitle(win_title)
        window.setWidget(widget)
        window.master = master
        real_values_at = self.pandas_dt[id].notna()
        y = self.pandas_dt[id][real_values_at].values
        x = self.pandas_dt["Timestamp (sec)"][real_values_at].values

        widget.slot_set_data(x=x, y=y)

        return window

    def tile_sub_windows(self):
        """
        Method to keep sub-windows tiled if input_keep_tiled checkbox is checked.
        Set tile mode based on self.view_mode variable.
        For some reason mdi area needs to be on top when reorganizing.
        That's why I'm scrolling to the top of the view first.
        :return:
        """
        if self.input_keep_tiled.isChecked():
            # Get width and height of mdi area
            w, h = self.mdi_area.width(), self.mdi_area.height()
            # Get all sub-windows in mdi area
            windows_list = self.mdi_area.subWindowList()
            if len(windows_list) > 0:
                # Scroll to the top of mdi area
                self.mdi_area.verticalScrollBar().setValue(0)
                self.mdi_area.horizontalScrollBar().setValue(0)
                if self.view_mode == MODE_LIST:
                    # List mode tiling
                    plot_height = None
                    if not self.input_view_scale_plots.isChecked():
                        plot_height = h * (self.input_view_plot_height.value() / 100)
                    self.tile_list_windows(list_windows=list(reversed(windows_list)), plot_height=plot_height)
                elif self.view_mode == MODE_ZOOM:
                    # Zoom mode tiling
                    master_height = h / 3.8
                    master_window = self.master_window
                    windows_list.remove(master_window)
                    master_window.setGeometry(0, h - master_height, w, master_height)
                    self.tile_slave_windows(slave_windows=windows_list, h=h - master_height)

    def sub_window_closed(self, event, window):
        """
        Sub Window closing event.
        Closed window needs to be removed from mdi area and selected rows as well.
        :param event:
        :param window:
        :return:
        """
        window.setParent(None)
        selection_model = self.input_header_items.selectionModel()
        model = self.input_header_items.model()
        row = model.model_data.index(window.id)
        deselection = QItemSelection(model.createIndex(row, row), model.createIndex(row, row))
        selection_model.select(deselection, QItemSelectionModel.Deselect | QItemSelectionModel.Rows)
        if self.view_tiled:
            self.slot_selection_changed()

    def master_zoom_changed(self, region):
        """
        Zoom changed in master window.
        :param region:
        :return:
        """
        self.last_zoom_region = region
        for window in self.mdi_area.subWindowList():
            if not window.master:
                window.widget().setXRange(*region, padding=0, update=False)
                window.widget().setClipToView(True)
                x1, x2 = int(region[0]), int(region[1])
                self.input_from_dt.setDateTime(QDateTime.fromTime_t(x1))
                self.input_to_dt.setDateTime(QDateTime.fromTime_t(x2))
                self.readout_diff_dt.setText(get_scaled_time_duration(x2 - x1))

    def add_sub_window(self, window):
        """
        Add sub window into mdi area.
        :param window:
        :return:
        """
        self.mdi_area.addSubWindow(window)
        if self.sub_windows_hint == Qt.FramelessWindowHint:
            window.widget().setTitle(window.windowTitle())
        else:
            window.widget().setTitle(None)
        window.show()

    def tile_slave_windows(self, slave_windows=None, x=0, y=0, w=None, h=None):
        """
        Organize set of slave windows in tile fashion.
        :param slave_windows: list
        :param x: int
        :param y: int
        :param w: int
        :param h: int
        :return: None
        """
        if slave_windows is None:
            slave_windows = self.mdi_area.subWindowList()
        if len(slave_windows) == 0:
            return
        if w is None:
            w = self.mdi_area.width()
        if h is None:
            h = self.mdi_area.height()
        for rows, i in enumerate(self.square_function_values):
            if i >= len(slave_windows):
                break
        try:
            if len(slave_windows) % rows:
                cols = int(len(slave_windows) / rows) + 1
            else:
                cols = int(len(slave_windows) / rows)
        except ZeroDivisionError:
            rows, cols = 1, 1
        slave_w = w / cols
        slave_h = h / rows
        w_pos, h_pos = 0, 0
        for index, window in enumerate(slave_windows):
            window.setGeometry(w_pos, h_pos, slave_w, slave_h)
            w_pos += slave_w
            if ((index + 1) % cols) == 0:
                h_pos += slave_h
                w_pos = 0

    def tile_list_windows(self, list_windows=None, plot_height=None):
        """
        Organize list windows in tile fashion.
        :param list_windows: list
        :param plot_height: int
        :return: None
        """
        if list_windows is None and len():
            list_windows = self.mdi_area.subWindowList()
        rows = len(list_windows)
        if rows == 0:
            return
        h = self.mdi_area.height()
        if plot_height is None:
            plot_height = h / rows

        w_offset = 0
        if plot_height * rows > h:
            w_offset = self.mdi_area.verticalScrollBar().width()

        w_width = self.mdi_area.width() - w_offset
        w_height = plot_height
        pos_w, pos_h = 0, 0
        for index, window in enumerate(list_windows):
            window.setGeometry(pos_w, pos_h, w_width, w_height)
            pos_h += w_height

    def resizeEvent(self, QResizeEvent):
        """
        Tile windows on resize event.
        :param QResizeEvent:
        :return:
        """
        self.tile_sub_windows()
        QWidget.resizeEvent(self, QResizeEvent)

    @pyqtSlot(QMdiSubWindow)
    def slot_window_activated(self, window):
        """
        Sub window was activated, perform following actions.
        :param window:
        :return:
        """
        follow_zoom_flag = self.input_zoom_follow.isChecked()
        if window is not None and not window.master and self.view_mode == MODE_ZOOM and follow_zoom_flag:
            self.input_master_row.setCurrentText(window.id)

    @pyqtSlot(bool)
    def slot_set_zoom_mode(self, flag):
        """
        View was changed zoom.
        Do nothing if flag is False.
        :param flag: bool
        :return: None
        """
        if flag:
            self.logger.info("View set to zoom mode.")
            self.view_mode = MODE_ZOOM
            self.tabs_view_tools.setCurrentIndex(0)
            current_sub_window = self.mdi_area.currentSubWindow()
            if current_sub_window is None:
                return
            zoom_master = self.input_master_row.currentText()
            if current_sub_window is not None:
                zoom_master = current_sub_window.id
            self.input_master_row.blockSignals(True)
            self.input_master_row.setCurrentText(zoom_master)
            self.input_master_row.blockSignals(False)
            self.slot_master_row_changed(zoom_master)

    @pyqtSlot(bool)
    def slot_set_list_mode(self, flag):
        """
        View was changed list.
        Do nothing if flag is False.
        :param flag: bool
        :return: None
        """
        if flag:
            self.logger.info("View set to list mode.")
            self.view_mode = MODE_LIST
            if self.master_window is not None:
                mater_window_id = self.master_window.id
                self.master_window.close()
                self.master_window.setParent(None)
                self.master_window = None
                for w in self.mdi_area.subWindowList():
                    if w.id == mater_window_id:
                        self.mdi_area.setActiveSubWindow(w)
                        break
            self.tabs_view_tools.setCurrentIndex(1)
            self.tile_sub_windows()

    @pyqtSlot(int, int)
    def slot_splitter_moved(self, x, y):
        """
        Layout splitter was moved.
        Tile all sub windows.
        :param x: int
        :param y: int
        :return: None
        """
        self.tile_sub_windows()

    @pyqtSlot(int)
    def slot_set_max_selected_rows(self, rows):
        """
        Maximum number of rows was changed.
        :param rows: int
        :return: None
        """
        self.logger.info("Maximum number of rows was set to: {}".format(rows))
        self.input_header_items.selectionModel().max_selected_rows = rows

    @pyqtSlot(str)
    def slot_master_row_changed(self, window_id):
        """
        Master row was changed.
        :param row_title: str
        :return:
        """
        if self.view_mode == MODE_ZOOM:
            self.mdi_area.blockSignals(True)
            last_zoom_region = self.last_zoom_region
            master_window = self.master_window
            if master_window is not None:
                master_window.close()
                master_window.setParent(None)
            sub_window = self.create_sub_window(window_id, master=True)
            self.add_sub_window(sub_window)
            self.tile_sub_windows()

            if last_zoom_region is not None:
                self.master_window.widget().lookingGlass.setRegion(last_zoom_region)
            self.mdi_area.blockSignals(False)

    @pyqtSlot()
    def slot_set_list_min_height(self):
        """
        Value of minimum plot height was changed.
        :return:
        """
        self.tile_sub_windows()

    @pyqtSlot()
    def slot_reset_zoom(self):
        """
        Reset zoom of Master plot.
        :return:
        """
        self.logger.info("Resetting zoom.")
        if self.master_window is not None:
            self.master_window.widget().reset_zoom()

    @pyqtSlot()
    def slot_reset_all(self):
        """
        Reset zoom of Master plot.
        :return:
        """
        self.logger.info("Resetting all plots, zoom and views.")
        self.slot_set_list_mode(True)
        self.master_window = None
        self.last_zoom_region = None
        self.view_tiled = True
        self.button_list_view.click()
        self.pushButton.click()
        self.pushButton_2.click()
        self.input_master_row.clear()

    @pyqtSlot()
    def slot_selection_changed(self, row_model_index=None):
        """
        Selection in rows list was changed.
        :param row_model_index: QItemSelection
        :return: None
        """
        self.view_tiled = False
        active_windows = []
        for w in self.mdi_area.subWindowList():
            if not w.master:
                active_windows.append(w.id)
        selection_model = self.input_header_items.selectionModel()
        if selection_model is not None:
            selected_row_indexes = selection_model.selectedIndexes()
            selected_rows = [str(row_index.data()) for row_index in selected_row_indexes]

            for window in self.mdi_area.subWindowList():
                if (window.id not in selected_rows) and not window.master:
                    window.close()
                    window.setParent(None)

            for row in selected_rows:
                if row not in active_windows:
                    self.add_sub_window(self.create_sub_window(row))

        self.tile_sub_windows()
        if self.view_mode == MODE_ZOOM:
            self.master_zoom_changed(self.master_window.widget().lookingGlass.getRegion())
        self.view_tiled = True

    def slot_select_rows(self, rows=None):
        if rows is None:
            return

        selection_model = self.input_header_items.selectionModel()
        model = self.input_header_items.model()
        for row_data in reversed(rows):
            index = model.model_data.index(row_data)
            if index > 0:
                selection = QItemSelection(model.createIndex(index, 0), model.createIndex(index, 0))
                selection_model.select(selection, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        self.slot_selection_changed()

    @pyqtSlot(bool)
    def slot_set_windowed_plots(self, flag):
        """
        Toggle windowed sub plots.
        Change sub windows hints and titles.
        :param flag: bool
        :return: None
        """
        self.input_keep_tiled.blockSignals(True)
        if not flag:
            self.input_keep_tiled.setEnabled(False)
            self.input_keep_tiled.setChecked(True)
        else:
            self.input_keep_tiled.setEnabled(True)
        self.input_keep_tiled.blockSignals(False)

        default_sub_window_hints = Qt.WindowSystemMenuHint | Qt.WindowCloseButtonHint | Qt.WindowMaximizeButtonHint
        self.sub_windows_hint = default_sub_window_hints if flag else Qt.FramelessWindowHint
        default_master_window_hint = Qt.CustomizeWindowHint | Qt.WindowTitleHint
        self.master_window_hint = default_master_window_hint if flag else Qt.FramelessWindowHint
        for window in self.mdi_area.subWindowList():
            if self.sub_windows_hint == Qt.FramelessWindowHint:
                window.widget().setTitle(window.windowTitle())
            else:
                window.widget().setTitle(None)

            if window.master and flag:
                window.setWindowFlags(self.master_window_hint)
            else:
                window.setWindowFlags(self.sub_windows_hint)

        self.tile_sub_windows()

    @pyqtSlot(bool)
    def slot_keep_tiled_changed(self, flag):
        """
        Toggle keep tiled changed.
        :param flag: bool
        :return: None
        """
        if flag:
            self.tile_sub_windows()

    def slot_export_to_pdf(self):
        if self.input_windowed_plots.isChecked():
            self.slot_set_windowed_plots(False)

        if len(self.mdi_area.subWindowList()) == 0:
            error_msg = "Export error: {}".format("Please select at least one plot to export.")
            self.logger.error(error_msg)
            show_error_message(error_msg, self)
            return

        plot_width, plot_height = self.input_plot_width.value(), self.input_plot_height.value()
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        extension_filter = "pdf [*.pdf] (*.pdf)"
        file_path, _ = QFileDialog.getSaveFileName(self, 'Set output file', "./{}".format("output.pdf"),
                                                   filter=extension_filter, initialFilter=extension_filter,
                                                   options=options)
        if file_path != "":
            self.logger.info("Exporting plots to {} using w: {}px, h: {}px".format(file_path, plot_width, plot_height))
            parts = []
            cache_dir = kamzik3.session.get_value(ATTR_CACHE_DIRECTORY)
            for index, window in enumerate(reversed(self.mdi_area.subWindowList())):
                actual_size = window.widget().size()
                window.widget().resize(plot_width, plot_height)
                exporter = exporters.ImageExporter(window.widget().plotItem)
                exporter.parameters()['width'] = plot_width
                img_path = os.path.join(cache_dir, 'img{}.png'.format(index))
                exporter.export(img_path)
                parts.append(Image(img_path))
                window.widget().resize(actual_size)

            while True:
                try:
                    use_parts = parts.copy()
                    report = SimpleDocTemplate(file_path, pagesize=(plot_width, plot_height))
                    report.build(use_parts)
                    break
                except LayoutError as e:
                    pw, ph = re.findall('([\-0-9\.]+)\sx\s([\-0-9\.]+)', str(e))[0]
                    plot_width += abs(plot_width - float(pw))
                    plot_height += abs(plot_height - float(ph))
                except Exception as e:
                    error_msg = "Export error: {}".format(e)
                    self.logger.error(error_msg)
                    show_error_message(error_msg, self)

            if self.input_windowed_plots.isChecked():
                self.slot_set_windowed_plots(True)

    def slot_export_to_csv(self):
        if len(self.mdi_area.subWindowList()) == 0:
            error_msg = "Export error: {}".format("Please select at least one plot to export.")
            self.logger.error(error_msg)
            show_error_message(error_msg, self)
            return

        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        extension_filter = "csv [*.csv] (*.csv)"
        file_path, _ = QFileDialog.getSaveFileName(self, 'Set output file', "./{}".format("output.csv"),
                                                   filter=extension_filter, initialFilter=extension_filter,
                                                   options=options)
        if file_path != "":
            self.logger.info("Exporting plots to {}".format(file_path))
            sel_mode = self.input_header_items.selectionModel()
            columns = []

            for item in sel_mode.selectedRows(0):
                columns.append(item.data())
            if self.last_zoom_region is None:
                self.pandas_dt.to_csv(file_path, columns=columns, index=False)
            else:
                head = self.pandas_dt.columns
                self.pandas_dt[(self.pandas_dt[head[1]] >= self.last_zoom_region[0]) & (
                            self.pandas_dt[head[1]] <= self.last_zoom_region[1])].to_csv(file_path, columns=columns,
                                                                                   index=False)



import datetime
import math

from pint import UndefinedUnitError

from kamzik3 import units
from kamzik3.constants import *


def _base_unit(unit):
    try:
        prefix, base_unit, suffix = tuple(units.parse_unit_name(unit))[0]
    except (IndexError, UndefinedUnitError, AttributeError, StopIteration):
        prefix, base_unit, suffix = None, unit, None
    return prefix, base_unit, suffix


def get_attribute_unit_range(attribute):
    prefix, base_unit, suffix = _base_unit(attribute[UNIT])
    if base_unit in ("ampere", "volt", "meter", "degree", "radian", "hertz", "percent"):
        sorted_prefixes = (
            'y', 'z', 'a', 'f', 'p', 'n', 'u', 'm', "", 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'
        )
    elif base_unit == "second":
        return ["ms", "s", "min"]
    elif base_unit in ("degree_Celsius", "kelvin", "degree_Fahrenheit"):
        return ["degC", "degF", "K"]
    elif base_unit in ("bar", "Pa", "atm"):
        return ["mbar", "bar", "Pa", "kPa", "atm"]
    elif base_unit == "revolutions_per_minute":
        return ["rpm", "rps"]
    else:
        sorted_prefixes = (
            'y', 'z', 'a', 'f', 'p', 'n', 'u', 'm', 'c', 'd', "", 'da', 'h', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'
        )
    try:
        converter = units._prefixes[prefix]._converter
        if converter == 1:
            scale = 1
        else:
            scale = converter.scale
    except KeyError:
        scale = 1

    try:
        min_unit = scale * 10 ** -attribute[DECIMALS]
        abs_min, abs_max = abs(attribute[MIN]), abs(attribute[MAX])
        max_unit = scale * (abs_max if abs_max > abs_min else abs_min)
    except TypeError:
        raise TypeError

    unit_range = []
    for prefix in sorted_prefixes:
        converter = units._prefixes[prefix]._converter
        if converter == 1:
            scale = 1
        else:
            scale = converter.scale
        if min_unit <= scale <= max_unit:
            unit_range.append(units.get_symbol("{}{}".format(prefix, base_unit)))

    return unit_range


def get_decimals_diff(old_unit, new_unit):
    old_scale = math.log(units.get_base_units(old_unit)[0], 10)
    new_scale = math.log(units.get_base_units(new_unit)[0], 10)
    return int(round(old_scale - new_scale))


def device_units(device, attribute, value):
    value = str(value)
    try:
        value = value.replace(u"%", u"percent")
    except AttributeError:
        pass
    converted_value = units.Quantity(value)

    device_unit = device.get_attribute(attribute)[UNIT]
    # try to convert value:

    # No unit for value
    if converted_value.unitless:
        # Return in value in device units
        try:
            pint_value = units.Quantity(value)
            if pint_value.u == units.dimensionless:
                return units.Quantity(pint_value.m, device_unit)
            else:
                return pint_value.to(device_unit)
        except AttributeError:
            return units.Quantity(float(units.Quantity(value).m))
    else:
        # Return converted value
        return converted_value.to(device_unit)


MAX_SECONDS = 3.154e+9


def get_scaled_time_duration(seconds, short=False):
    if seconds < 0:
        return
    if seconds >= MAX_SECONDS:
        return "> 100 Years"

    if not short:
        sec, msec, min, day, hour, month, year = "Sec", "MSec", "Min", "Day", "Hour", "Month", "Year"
    else:
        sec, msec, min, day, hour, month, year = "s", "ms", "m", "D", "h", "M", "Y"

    d = datetime.datetime(1, 1, 1) + datetime.timedelta(seconds=seconds)

    if seconds < 60:
        milliseconds = int(d.microsecond / 1000)
        if milliseconds > 0:
            time_estimate = "{} {} {} {}".format(d.second, sec, milliseconds, msec)
        else:
            time_estimate = "{} {}".format(d.second, sec)
    elif 60 <= seconds < 3600:
        time_estimate = "{} {} {} {}".format(d.minute, min, d.second, sec)
    elif 3600 <= seconds < 24 * 3600:
        time_estimate = "{} {} {} {}".format(d.hour, hour, d.minute, min)
    elif 24 * 3600 <= seconds < 24 * 3600 * 28:
        time_estimate = "{} {} {} {} {} {}".format(d.day - 1, day, d.hour, hour, d.minute, min)
    elif 24 * 3600 * 28 <= seconds < 24 * 3600 * 28 * 12:
        time_estimate = "{} {} {} {} {} {}".format(d.month - 1, month, d.day, day, d.hour, hour)
    else:
        time_estimate = "{} {} {} {} {} {}".format(d.year - 1, year, d.month, month, d.day, day)

    return time_estimate


def seconds_to_datetime(seconds):
    if seconds > MAX_SECONDS:
        seconds = MAX_SECONDS

    return datetime.datetime.fromtimestamp(seconds).strftime("%d.%m.%Y %H:%M:%S")


def get_printable_size(byte_size):
    """
    A bit is the smallest unit, it's either 0 or 1
    1 byte = 1 octet = 8 bits
    1 kB = 1 kilobyte = 1000 bytes = 10^3 bytes
    1 KiB = 1 kibibyte = 1024 bytes = 2^10 bytes
    1 KB = 1 kibibyte OR kilobyte ~= 1024 bytes ~= 2^10 bytes (it usually means 1024 bytes but sometimes it's 1000... ask the sysadmin ;) )
    1 kb = 1 kilobits = 1000 bits (this notation should not be used, as it is very confusing)
    1 ko = 1 kilooctet = 1000 octets = 1000 bytes = 1 kB

    Also Kb seems to be a mix of KB and kb, again it depends on context.

    In linux, a byte (B) is composed by a sequence of bits (b). One byte has 256 possible values.

    More info : http://www.linfo.org/byte.html
    """
    BASE_SIZE = 1024.00
    MEASURE = ["B", "KB", "MB", "GB", "TB", "PB"]

    def _fix_size(size, size_index):
        if not size:
            return "0"
        elif size_index == 0:
            return str(size)
        else:
            return "{:.3f}".format(size)

    current_size = byte_size
    size_index = 0

    while current_size >= BASE_SIZE and len(MEASURE) != size_index:
        current_size = current_size / BASE_SIZE
        size_index = size_index + 1

    size = _fix_size(current_size, size_index)
    measure = MEASURE[size_index]
    return size + " " + measure



import time
from threading import Event, Thread
import numpy as np
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsDecorators import expose_method


class DeviceGrabber(Device):

    def __init__(self, device_id=None, config=None):
        self.stopped = Event()
        self.frame_delay = 0
        Device.__init__(self, device_id, config)
        self.connect()

    def _init_attributes(self):
        Device._init_attributes(self)
        npa = np.zeros(dtype=np.uint8, shape=(1, 1, 3))
        self.create_attribute(ATTR_FRAME, group="Grabber", default_value=npa, default_type=TYPE_ARRAY,
                              unit="", readonly=True, display=IMAGE, min_broadcast_timeout=100)
        self.create_attribute(ATTR_MAX_BROADCAST_RATE, default_value=10, default_type=np.uint8, unit="Hz", min_value=0,
                              max_value=10000, set_function=self.set_max_broadcast_rate)
        self.create_attribute(ATTR_FRAMERATE, default_value=10, default_type=np.float, unit="FPS", min_value=1,
                              max_value=10000, set_function=self.set_framerate, decimals=2)
        self.create_attribute(ATTR_FRAMES_GRABBED, default_value=0, default_type=np.uint, readonly=True)

    @expose_method()
    def acquire(self):
        if self.get_value(ATTR_STATUS) == STATUS_BUSY:
            return
        Thread(target=self._acquire_loop).start()

    def _acquire_loop(self):
        """
        self.stopped.clear()
        self.frame_delay = 1. / self.get_value(ATTR_FRAMERATE)
        self.set_status(STATUS_BUSY)

        fps_counter_reference = time_reference = time.perf_counter_ns()
        grabbed_frames_count = 0
        fps_counter = 0
        while not self.stopped.wait((self.frame_delay - (time.perf_counter_ns() - time_reference)) * 1e-9):
            time_reference = time.perf_counter_ns()
            grab_frame_method
            grabbed_frames_count += 1
            fps_counter += 1
            if ret:
                self.set_value(["Grabber", "Frame"], cv2.cvtColor(cv_img, cv2.COLOR_BGR2RGB))

            if (time.perf_counter_ns() - fps_counter_reference) >= 1e9:
                self.set_raw_value(ATTR_FRAMERATE, fps_counter)
                self.set_raw_value(ATTR_FRAMES_GRABBED, grabbed_frames_count)
                fps_counter = 0
                fps_counter_reference = time.perf_counter_ns()
        """
        raise NotImplemented

    def set_max_broadcast_rate(self, value):
        self[ATTR_OUTPUT][ATTR_FRAME][ALLOW_BROADCAST] = value > 0
        if value > 0:
            self[ATTR_OUTPUT][ATTR_FRAME].min_broadcast_timeout = (1e9 / value) * 0.95

    def set_framerate(self, value):
        self.frame_delay = 1e9 / value

    def handle_configuration(self):
        start_at = time.time()
        self._config_attributes()
        self.set_status(STATUS_CONFIGURED)
        self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

    @expose_method()
    def stop(self):
        self.stopped.set()
        self.set_status(STATUS_IDLE)

    def close(self):
        self.stop()
        Device.close(self)
